import numpy as np
from Orange.classification.utils.fasterrisk.fasterrisk import (
    RiskScoreOptimizer,
    RiskScoreClassifier,
)

from Orange.classification import Learner, Model
from Orange.data import Table, Storage
from Orange.data.filter import HasClass
from Orange.preprocess import Discretize, Impute, Continuize, SelectBestFeatures
from Orange.preprocess.discretize import Binning
from Orange.preprocess.score import ReliefF


def _change_class_var_values(y):
    """
    Changes the class variable values from 0 and 1 to -1 and 1 or vice versa.
    """
    return np.where(y == 0, -1, np.where(y == -1, 0, y))


class ScoringSheetModel(Model):
    def __init__(self, model):
        self.model = model
        super().__init__()

    def predict_storage(self, table):
        if not isinstance(table, Storage):
            raise TypeError("Data is not a subclass of Orange.data.Storage.")

        y_pred = _change_class_var_values(self.model.predict(table.X))
        y_prob = self.model.predict_prob(table.X)

        scores = np.hstack(((1 - y_prob).reshape(-1, 1), y_prob.reshape(-1, 1)))
        return y_pred, scores


class ScoringSheetLearner(Learner):
    __returns__ = ScoringSheetModel
    preprocessors = [HasClass(), Discretize(method=Binning()), Impute(), Continuize()]

    def __init__(
        self,
        num_attr_after_selection=20,
        num_decision_params=5,
        max_points_per_param=5,
        num_input_features=None,
        preprocessors=None,
    ):
        # Set the num_decision_params, max_points_per_param, and num_input_features normally
        self.num_decision_params = num_decision_params
        self.max_points_per_param = max_points_per_param
        self.num_input_features = num_input_features
        self.feature_to_group = None

        if preprocessors is None:
            self.preprocessors = [
                *self.preprocessors,
                SelectBestFeatures(method=ReliefF(), k=num_attr_after_selection),
            ]

        super().__init__(preprocessors=preprocessors)

    def incompatibility_reason(self, domain):
        reason = None
        if len(domain.class_vars) > 1 and not self.supports_multiclass:
            reason = "Too many target variables."
        elif not domain.has_discrete_class:
            reason = "Categorical class variable expected."
        elif len(domain.class_vars[0].values) > 2:
            reason = "Too many target variable values."
        return reason

    def fit_storage(self, table):
        if not isinstance(table, Storage):
            raise TypeError("Data is not a subclass of Orange.data.Storage.")
        elif table.get_nan_count_class() > 0:
            raise ValueError("Class variable contains missing values.")

        if self.num_input_features is not None:
            self._generate_feature_group_index(table)

        X, y, _ = table.X, table.Y, table.W if table.has_weights() else None
        learner = RiskScoreOptimizer(
            X=X,
            y=_change_class_var_values(y),
            k=self.num_decision_params,
            select_top_m=1,
            lb=-self.max_points_per_param,
            ub=self.max_points_per_param,
            group_sparsity=self.num_input_features,
            featureIndex_to_groupIndex=self.feature_to_group,
        )

        self._optimize_decision_params_adjustment(learner)

        multipliers, intercepts, coefficients = learner.get_models()

        model = RiskScoreClassifier(
            multiplier=multipliers[0],
            intercept=intercepts[0],
            coefficients=coefficients[0],
            featureNames=[attribute.name for attribute in table.domain.attributes],
            X_train=X if self.num_decision_params > 10 else None,
        )

        return ScoringSheetModel(model)

    def _optimize_decision_params_adjustment(self, learner):
        """
        This function attempts to optimize (fit) the learner, reducing the number of decision
        parameters ('k')if optimization fails due to being too high.

        Sometimes, the number of decision parameters is too high for the
        number of input features. Which results in a ValueError.
        Continues until successful or 'k' cannot be reduced further.
        """
        while True:
            try:
                learner.optimize()
                return True
            except ValueError as e:
                learner.k -= 1
                if learner.k < 1:
                    # Raise a custom error when k falls below 1
                    raise ValueError(
                        "The number of input features is too low for the current settings."
                    ) from e

    def _generate_feature_group_index(self, table):
        """
        Returns a feature index to group index mapping. The group index is used to group
        binarized features that belong to the same original feature.
        """
        original_feature_names = [
            attribute.compute_value.variable.name
            for attribute in table.domain.attributes
        ]
        feature_to_group_index = {
            feature: idx for idx, feature in enumerate(set(original_feature_names))
        }
        feature_to_group = [
            feature_to_group_index[feature] for feature in original_feature_names
        ]
        self.feature_to_group = np.asarray(feature_to_group)


if __name__ == "__main__":
    mock_learner = ScoringSheetLearner(20, 5, 10, None)
    mock_table = Table("https://datasets.biolab.si/core/heart_disease.tab")
    mock_model = mock_learner(mock_table)
    mock_model(mock_table)



from AnyQt.QtCore import Qt

from Orange.data import Table
from Orange.base import Model
from Orange.widgets.utils.owlearnerwidget import OWBaseLearner
from Orange.widgets.utils.concurrent import TaskState, ConcurrentWidgetMixin
from Orange.widgets.widget import Msg
from Orange.widgets import gui
from Orange.widgets.settings import Setting

from Orange.classification.scoringsheet import ScoringSheetLearner


class ScoringSheetRunner:
    @staticmethod
    def run(learner: ScoringSheetLearner, data: Table, state: TaskState) -> Model:
        if data is None:
            return None
        state.set_status("Learning...")
        model = learner(data)
        return model


class OWScoringSheet(OWBaseLearner, ConcurrentWidgetMixin):
    name = "Scoring Sheet"
    description = "A fast and explainable classifier."
    icon = "icons/ScoringSheet.svg"
    replaces = ["orangecontrib.prototypes.widgets.owscoringsheet.OWScoringSheet"]
    priority = 75
    keywords = "scoring sheet"

    LEARNER = ScoringSheetLearner

    class Inputs(OWBaseLearner.Inputs):
        pass

    class Outputs(OWBaseLearner.Outputs):
        pass

    # Preprocessing
    num_attr_after_selection = Setting(20)

    # Scoring Sheet Settings
    num_decision_params = Setting(5)
    max_points_per_param = Setting(5)
    custom_features_checkbox = Setting(False)
    num_input_features = Setting(1)

    # Warning messages
    class Information(OWBaseLearner.Information):
        custom_num_of_input_features = Msg(
            "If the number of input features used is too low for the number of decision \n"
            "parameters, the number of decision parameters will be adjusted to fit the model."
        )

    def __init__(self):
        ConcurrentWidgetMixin.__init__(self)
        OWBaseLearner.__init__(self)

    def add_main_layout(self):
        box = gui.vBox(self.controlArea, "Preprocessing")

        self.num_attr_after_selection_spin = gui.spin(
            box,
            self,
            "num_attr_after_selection",
            minv=1,
            maxv=100,
            step=1,
            label="Number of Attributes After Feature Selection:",
            orientation=Qt.Horizontal,
            alignment=Qt.AlignRight,
            callback=self.settings_changed,
            controlWidth=45,
        )

        box = gui.vBox(self.controlArea, "Model Parameters")

        gui.spin(
            box,
            self,
            "num_decision_params",
            minv=1,
            maxv=50,
            step=1,
            label="Maximum Number of Decision Parameters:",
            orientation=Qt.Horizontal,
            alignment=Qt.AlignRight,
            callback=self.settings_changed,
            controlWidth=45,
        )

        gui.spin(
            box,
            self,
            "max_points_per_param",
            minv=1,
            maxv=100,
            step=1,
            label="Maximum Points per Decision Parameter:",
            orientation=Qt.Horizontal,
            alignment=Qt.AlignRight,
            callback=self.settings_changed,
            controlWidth=45,
        )

        gui.checkBox(
            box,
            self,
            "custom_features_checkbox",
            label="Custom number of input features",
            callback=[self.settings_changed, self.custom_input_features],
        )

        self.custom_features = gui.spin(
            box,
            self,
            "num_input_features",
            minv=1,
            maxv=50,
            step=1,
            label="Number of Input Features Used:",
            orientation=Qt.Horizontal,
            alignment=Qt.AlignRight,
            callback=self.settings_changed,
            controlWidth=45,
        )

        self.custom_input_features()

    def custom_input_features(self):
        self.custom_features.setEnabled(self.custom_features_checkbox)
        if self.custom_features_checkbox:
            self.Information.custom_num_of_input_features()
        else:
            self.Information.custom_num_of_input_features.clear()
        self.apply()

    @Inputs.data
    def set_data(self, data):
        self.cancel()
        super().set_data(data)

    @Inputs.preprocessor
    def set_preprocessor(self, preprocessor):
        self.cancel()
        super().set_preprocessor(preprocessor)

        # Enable or disable the spin box based on whether a preprocessor is set
        self.num_attr_after_selection_spin.setEnabled(preprocessor is None)
        if preprocessor:
            self.Information.ignored_preprocessors()
        else:
            self.Information.ignored_preprocessors.clear()

    def create_learner(self):
        return self.LEARNER(
            num_attr_after_selection=self.num_attr_after_selection,
            num_decision_params=self.num_decision_params,
            max_points_per_param=self.max_points_per_param,
            num_input_features=(
                self.num_input_features if self.custom_features_checkbox else None
            ),
            preprocessors=self.preprocessors,
        )

    def update_model(self):
        self.cancel()
        self.show_fitting_failed(None)
        self.model = None
        if self.data is not None:
            self.start(ScoringSheetRunner.run, self.learner, self.data)
        else:
            self.Outputs.model.send(None)

    def get_learner_parameters(self):
        return (
            self.num_decision_params,
            self.max_points_per_param,
            self.num_input_features,
        )

    def on_partial_result(self, _):
        pass

    def on_done(self, result: Model):
        assert isinstance(result, Model) or result is None
        self.model = result
        self.Outputs.model.send(result)

    def on_exception(self, ex):
        self.cancel()
        self.Outputs.model.send(None)
        if isinstance(ex, BaseException):
            self.show_fitting_failed(ex)

    def onDeleteWidget(self):
        self.shutdown()
        super().onDeleteWidget()


if __name__ == "__main__":
    from Orange.widgets.utils.widgetpreview import WidgetPreview

    WidgetPreview(OWScoringSheet).run()



import unittest

from AnyQt.QtCore import Qt

from orangewidget.tests.base import WidgetTest

from Orange.data import Table
from Orange.classification import LogisticRegressionLearner, ScoringSheetLearner
from Orange.widgets.widget import AttributeList
from Orange.widgets.visualize.owscoringsheetviewer import OWScoringSheetViewer


class TestOWScoringSheetViewer(WidgetTest):
    @classmethod
    def setUpClass(cls):
        super().setUpClass()
        cls.heart = Table("heart_disease")
        cls.scoring_sheet_learner = ScoringSheetLearner(20, 5, 5, None)
        cls.scoring_sheet_model = cls.scoring_sheet_learner(cls.heart)
        cls.logistic_regression_learner = LogisticRegressionLearner(tol=1)
        cls.logistic_regression_model = cls.logistic_regression_learner(cls.heart[:10])

    def setUp(self):
        self.widget = self.create_widget(OWScoringSheetViewer)

    def test_no_classifier_input(self):
        coef_table = self.widget.coefficient_table
        risk_slider = self.widget.risk_slider
        class_combo = self.widget.class_combo

        self.assertEqual(coef_table.rowCount(), 0)
        self.assertEqual(risk_slider.slider.value(), 0)
        self.assertEqual(class_combo.count(), 0)

    def test_no_classifier_output(self):
        self.assertIsNone(self.get_output(self.widget.Outputs.features))
        self.send_signal(self.widget.Inputs.classifier, self.scoring_sheet_model)
        self.send_signal(self.widget.Inputs.classifier, None)
        self.assertIsNone(self.get_output(self.widget.Outputs.features))

    def test_classifier_output(self):
        self.send_signal(self.widget.Inputs.classifier, self.scoring_sheet_model)
        output = self.get_output(self.widget.Outputs.features)
        self.assertIsInstance(output, AttributeList)
        self.assertEqual(len(output), self.scoring_sheet_learner.num_decision_params)

    def test_table_population_on_model_input(self):
        self.send_signal(self.widget.Inputs.classifier, self.scoring_sheet_model)
        table = self.widget.coefficient_table
        self.assertEqual(
            table.rowCount(), self.scoring_sheet_learner.num_decision_params
        )

        for column in range(table.columnCount()):
            for row in range(table.rowCount()):
                self.assertIsNotNone(table.item(row, column))
                if column == 2:
                    self.assertEqual(table.item(row, column).checkState(), Qt.Unchecked)

    def test_slider_population_on_model_input(self):
        self.send_signal(self.widget.Inputs.classifier, self.scoring_sheet_model)
        slider = self.widget.risk_slider
        self.assertIsNotNone(slider.points)
        self.assertIsNotNone(slider.probabilities)
        self.assertEqual(len(slider.points), len(slider.probabilities))

    def test_slider_update_on_checkbox_toggle(self):
        self.send_signal(self.widget.Inputs.classifier, self.scoring_sheet_model)

        coef_table = self.widget.coefficient_table
        risk_slider = self.widget.risk_slider
        risk_slider_points = risk_slider.points

        # Get the items in the first row of the table
        checkbox_item = coef_table.item(0, 2)
        attribute_points_item = coef_table.item(0, 1)

        # Check if the slider value is "0" before changing the checkbox
        self.assertEqual(risk_slider.slider.value(), risk_slider_points.index(0))

        # Directly change the checkbox state to Checked
        checkbox_item.setCheckState(Qt.Checked)

        # Re-fetch the items after change
        attribute_points_item = coef_table.item(0, 1)

        # Check if the slider value is now the same as the attribute's coefficient
        self.assertEqual(
            risk_slider.slider.value(),
            risk_slider_points.index(float(attribute_points_item.text())),
        )

        # Directly change the checkbox state to Unchecked
        checkbox_item.setCheckState(Qt.Unchecked)

        # Check if the slider value is "0" again
        self.assertEqual(risk_slider.slider.value(), risk_slider_points.index(0))

    def test_target_class_change(self):
        self.send_signal(self.widget.Inputs.classifier, self.scoring_sheet_model)
        self.class_combo = self.widget.class_combo

        # Check if the values of the combobox "match" the domain
        self.assertEqual(
            self.class_combo.count(),
            len(self.scoring_sheet_model.domain.class_var.values),
        )
        for i in range(self.class_combo.count()):
            self.assertEqual(
                self.class_combo.itemText(i),
                self.scoring_sheet_model.domain.class_var.values[i],
            )

        old_coefficients = self.widget.coefficients.copy()
        old_all_scores = self.widget.all_scores.copy()
        old_all_risks = self.widget.all_risks.copy()

        # Change the target class to the second class
        self.class_combo.setCurrentIndex(1)
        self.widget._class_combo_changed()

        # Check if the coefficients, scores, and risks have changed
        self.assertNotEqual(old_coefficients, self.widget.coefficients)
        self.assertNotEqual(old_all_scores, self.widget.all_scores)
        self.assertNotEqual(old_all_risks, self.widget.all_risks)

    def test_invalid_classifier_error(self):
        self.send_signal(self.widget.Inputs.classifier, self.logistic_regression_model)
        self.assertTrue(self.widget.Error.invalid_classifier.is_shown())
        self.send_signal(self.widget.Inputs.classifier, self.scoring_sheet_model)
        self.assertFalse(self.widget.Error.invalid_classifier.is_shown())

    def test_multiple_instances_information(self):
        self.send_signal(self.widget.Inputs.data, self.heart[:2])
        self.assertTrue(self.widget.Information.multiple_instances.is_shown())
        self.send_signal(self.widget.Inputs.data, self.heart[:1])
        self.assertFalse(self.widget.Information.multiple_instances.is_shown())

    def _get_checkbox_states(self, coef_table):
        for row in range(coef_table.rowCount()):
            if self.widget.instance_points[row] == 1:
                self.assertEqual(coef_table.item(row, 2).checkState(), Qt.Checked)
            else:
                self.assertEqual(coef_table.item(row, 2).checkState(), Qt.Unchecked)

    def test_checkbox_after_instance_input(self):
        self.send_signal(self.widget.Inputs.classifier, self.scoring_sheet_model)
        self.send_signal(self.widget.Inputs.data, self.heart[:1])
        coef_table = self.widget.coefficient_table
        self._get_checkbox_states(coef_table)
        self.send_signal(self.widget.Inputs.data, self.heart[1:2])
        self._get_checkbox_states(coef_table)

    def test_no_classifier_UI(self):
        coef_table = self.widget.coefficient_table
        risk_slider = self.widget.risk_slider
        class_combo = self.widget.class_combo

        self.assertEqual(coef_table.rowCount(), 0)
        self.assertEqual(risk_slider.points, [])
        self.assertEqual(class_combo.count(), 0)

        self.send_signal(self.widget.Inputs.classifier, self.scoring_sheet_model)

        self.assertEqual(
            coef_table.rowCount(), self.scoring_sheet_learner.num_decision_params
        )
        self.assertIsNotNone(risk_slider.points)
        self.assertEqual(
            class_combo.count(), len(self.scoring_sheet_model.domain.class_var.values)
        )

        self.send_signal(self.widget.Inputs.classifier, None)

        self.assertEqual(coef_table.rowCount(), 0)
        self.assertEqual(risk_slider.points, [])
        self.assertEqual(class_combo.count(), 0)


if __name__ == "__main__":
    unittest.main()



# pylint: disable=protected-access
import unittest
from unittest.mock import Mock

from orangewidget.tests.base import GuiTest

from Orange.data import Table
from Orange.widgets.visualize.utils.error_bars_dialog import ErrorBarsDialog


class TestErrorBarsDialog(GuiTest):
    def setUp(self) -> None:
        self._dlg = ErrorBarsDialog(None)

    def test_init(self):
        form = self._dlg.layout().itemAt(0)

        self.assertEqual(form.itemAt(0).widget().text(), "Upper:")
        self.assertEqual(form.itemAt(1).widget().currentText(), "(None)")

        self.assertEqual(form.itemAt(2).widget().text(), "Lower:")
        self.assertEqual(form.itemAt(3).widget().currentText(), "(None)")

        self.assertEqual(form.itemAt(4).widget().text(),
                         "Difference from plotted value")
        self.assertTrue(form.itemAt(4).widget().isChecked())

        self.assertEqual(form.itemAt(5).widget().text(),
                         "Absolute position on the plot")
        self.assertFalse(form.itemAt(5).widget().isChecked())

    def test_get_data(self):
        upper_var, lower_var, is_abs = self._dlg.get_data()
        self.assertIsNone(upper_var)
        self.assertIsNone(lower_var)
        self.assertFalse(is_abs)

    def test_set_data(self):
        data = Table("iris")

        self._dlg._set_data(data.domain, data.domain.attributes[2],
                            data.domain.attributes[1], True)
        upper_var, lower_var, is_abs = self._dlg.get_data()
        self.assertIs(upper_var, data.domain.attributes[2])
        self.assertIs(lower_var, data.domain.attributes[1])
        self.assertTrue(is_abs)

        self._dlg._set_data(data.domain, None, None, True)
        upper_var, lower_var, is_abs = self._dlg.get_data()
        self.assertIsNone(upper_var)
        self.assertIsNone(lower_var)
        self.assertTrue(is_abs)

    def test_set_data_none(self):
        self._dlg._set_data(None, None, None, False)
        upper_var, lower_var, is_abs = self._dlg.get_data()
        self.assertIsNone(upper_var)
        self.assertIsNone(lower_var)
        self.assertFalse(is_abs)

    def test_set_data_err(self):
        data = Table("iris")
        self.assertRaises(ValueError, self._dlg._set_data, data.domain,
                          data.domain.class_var, data.domain.class_var, False)

    def test_changed(self):
        data = Table("iris")
        mock = Mock()
        self._dlg.changed.connect(mock)
        self._dlg._set_data(data.domain, data.domain.attributes[2],
                            data.domain.attributes[1], True)

        self._dlg._ErrorBarsDialog__upper_combo.setCurrentIndex(1)
        mock.assert_called_once()

        mock.reset_mock()
        self._dlg._ErrorBarsDialog__lower_combo.setCurrentIndex(0)
        mock.assert_called_once()

        mock.reset_mock()
        self._dlg._ErrorBarsDialog__radio_buttons.buttons()[0].click()
        mock.assert_called_once()


if __name__ == "__main__":
    unittest.main()



import time
import unittest
from unittest.mock import patch, Mock

import requests

from AnyQt.QtCore import QItemSelectionModel, Qt

from Orange.widgets.data.owdatasets import OWDataSets, Namespace as DSNamespace, \
    GENERAL_DOMAIN, ALL_DOMAINS
from Orange.widgets.tests.base import WidgetTest


class TestOWDataSets(WidgetTest):
    def setUp(self):
        # Most tests check the iniitialization of widget under different
        # conditions, therefore mocks are needed prior to calling createWidget.
        # Inherited methods will set self.widget; here we set it to None to
        # avoid lint errors.
        self.widget = None

    @patch("Orange.widgets.data.owdatasets.list_remote",
           Mock(side_effect=requests.exceptions.ConnectionError))
    @patch("Orange.widgets.data.owdatasets.list_local",
           Mock(return_value={}))
    @patch("Orange.widgets.data.owdatasets.log", Mock())
    def test_no_internet_connection(self):
        w = self.create_widget(OWDataSets)  # type: OWDataSets
        self.wait_until_stop_blocking(w)
        self.assertTrue(w.Error.no_remote_datasets.is_shown())

    @patch("Orange.widgets.data.owdatasets.list_remote",
           Mock(side_effect=requests.exceptions.ConnectionError))
    @patch("Orange.widgets.data.owdatasets.list_local",
           Mock(return_value={('core', 'foo.tab'): {}}))
    @patch("Orange.widgets.data.owdatasets.log", Mock())
    @WidgetTest.skipNonEnglish
    def test_only_local(self):
        w = self.create_widget(OWDataSets)  # type: OWDataSets
        self.wait_until_stop_blocking(w)
        self.assertTrue(w.Warning.only_local_datasets.is_shown())
        self.assertEqual(w.view.model().rowCount(), 1)

    @patch("Orange.widgets.data.owdatasets.list_remote",
           Mock(side_effect=requests.exceptions.ConnectionError))
    @patch("Orange.widgets.data.owdatasets.list_local",
           Mock(return_value={('core', 'foo.tab'): {"language": "English"},
                              ('core', 'bar.tab'): {"language": "Slovenščina"}}))
    @patch("Orange.widgets.data.owdatasets.log", Mock())
    def test_filtering(self):
        w = self.create_widget(OWDataSets)  # type: OWDataSets
        model = w.view.model()
        model.setLanguage(None)
        self.wait_until_stop_blocking(w)
        self.assertEqual(model.rowCount(), 2)
        w.filterLineEdit.setText("foo")
        self.assertEqual(model.rowCount(), 1)
        w.filterLineEdit.setText("baz")
        self.assertEqual(model.rowCount(), 0)
        w.filterLineEdit.setText("")
        self.assertEqual(model.rowCount(), 2)

        model.setLanguage("Slovenščina")
        self.assertEqual(model.rowCount(), 1)
        self.assertEqual(model.index(0, 0).data(Qt.UserRole).title, "bar.tab")

        model.setLanguage("English")
        self.assertEqual(model.rowCount(), 1)
        self.assertEqual(model.index(0, 0).data(Qt.UserRole).title, "foo.tab")

        model.setLanguage(None)
        self.assertEqual(model.rowCount(), 2)

    @patch("Orange.widgets.data.owdatasets.list_remote",
           Mock(side_effect=requests.exceptions.ConnectionError))
    @patch("Orange.widgets.data.owdatasets.list_local",
           Mock(return_value={('core', 'foo.tab'): {"domain": None},
                              ('edu', 'bar.tab'): {"domain": "edu"}}))
    @patch("Orange.widgets.data.owdatasets.log", Mock())
    def test_filtering_by_domain(self):
        w = self.create_widget(OWDataSets)  # type: OWDataSets
        model = w.view.model()
        model.setDomain(GENERAL_DOMAIN)
        self.wait_until_stop_blocking(w)
        self.assertEqual(model.rowCount(), 1)

        model.setDomain(ALL_DOMAINS)
        self.wait_until_stop_blocking(w)
        self.assertEqual(model.rowCount(), 2)

        model.setDomain("edu")
        self.assertEqual(model.rowCount(), 1)
        self.assertEqual(model.index(0, 0).data(Qt.UserRole).title, "bar.tab")

        model.setDomain("baz")
        self.assertEqual(model.rowCount(), 0)

    @patch("Orange.widgets.data.owdatasets.list_local",
           Mock(return_value={('core', 'foo.tab'): {"domain": None},
                              ('core', 'bar.tab'): {"domain": "edu"}}))
    @patch("Orange.widgets.data.owdatasets.log", Mock())
    @patch("Orange.widgets.data.owdatasets.OWDataSets.commit", Mock())
    def test_change_domain(self):
        def wait_and_return(_):
            time.sleep(0.2)
            return {('core', 'foo.tab'): {"domain": "edu"},
                    ('core', 'bar.tab'): {"domain": "edu"}}
        with patch("Orange.widgets.data.owdatasets.list_remote",
                   new=wait_and_return):
            self.widget = w = self.create_widget(OWDataSets,
                                   stored_settings={"selected_id": "bar.tab",
                                                    "domain": "edu"})
            self.wait_until_stop_blocking()
            self.assertEqual(w.selected_id, "bar.tab")
            self.assertEqual(w.domain_combo.currentText(), "edu")

            self.widget = w = self.create_widget(OWDataSets,
                                   stored_settings={"selected_id": "foo.tab",
                                                    "domain": "(core)"})
            self.wait_until_stop_blocking()
            self.assertEqual(w.selected_id, "foo.tab")
            self.assertEqual(w.domain_combo.currentText(), "edu")

            self.widget = w = self.create_widget(OWDataSets,
                                   stored_settings={"selected_id": "bar.tab",
                                                    "domain": "(core)"})
            self.wait_until_stop_blocking()
            self.assertEqual(w.selected_id, "bar.tab")
            self.assertEqual(w.domain_combo.currentText(), "edu")

    def __titles(self, widget):
        model = widget.view.model()
        return {
            model.index(row, 0).data(Qt.UserRole).title
            for row in range(model.rowCount())}

    @patch("Orange.widgets.data.owdatasets.list_remote",
           Mock(side_effect=requests.exceptions.ConnectionError))
    @patch("Orange.widgets.data.owdatasets.list_local",
           Mock(return_value={
               ('core', 'foo.tab'): {"title": "an unlisted data set",
                                     "publication_status": DSNamespace.UNLISTED},
               ('core', 'bar.tab'): {"title": "a published data set",
                                     "publication_status": DSNamespace.PUBLISHED},
               ('core', 'baz.tab'): {"title": "an unp unp",
                                     "publication_status": DSNamespace.PUBLISHED}
           }))
    @patch("Orange.widgets.data.owdatasets.log", Mock())
    def test_filtering_unlisted(self):
        def titles():
            return self.__titles(w)

        w = self.create_widget(OWDataSets)  # type: OWDataSets
        model = w.view.model()
        self.assertEqual(titles(), {"a published data set", "an unp unp"})

        model.setFilterFixedString("unp")
        self.assertEqual(titles(), {"an unp unp"})

        model.setFilterFixedString("an U")
        self.assertEqual(titles(), {"an unlisted data set", "an unp unp"})

        model.setFilterFixedString("")
        self.assertEqual(titles(), {"a published data set", "an unp unp"})

        model.setFilterFixedString(None)
        self.assertEqual(titles(), {"a published data set", "an unp unp"})

    @patch("Orange.widgets.data.owdatasets.list_remote",
           Mock(return_value={('core', 'foo.tab'): {"title": "Foo data set",
                                                    "language": "English"},
                              ('core', 'bar.tab'): {"title": "Bar data set",
                                                    "domain": "Testing"},
                              ('core', 'bax.tab'): {"title": "Bax data set",
                                                    "language": "Slovenščina"}
                              }))
    @patch("Orange.widgets.data.owdatasets.list_local",
           Mock(return_value={}))
    @patch("Orange.widgets.data.owdatasets.OWDataSets.commit", Mock())
    def test_filter_overrides_language_and_domain(self):
        w = self.create_widget(OWDataSets)  # type: OWDataSets
        self.wait_until_stop_blocking(w)
        w.language_combo.setCurrentText("Slovenščina")
        w.language_combo.activated.emit(w.language_combo.currentIndex())
        w.domain_combo.setCurrentText(w.GENERAL_DOMAIN_LABEL)
        w.domain_combo.activated.emit(w.domain_combo.currentIndex())

        self.assertEqual(self.__titles(w), {"Bax data set"})

        w.filterLineEdit.setText("data ")
        self.assertEqual(self.__titles(w), {"Foo data set",
                                            "Bar data set",
                                            "Bax data set"})
        self.assertEqual(w.language_combo.currentText(), w.ALL_LANGUAGES)
        self.assertFalse(w.language_combo.isEnabled())
        self.assertEqual(w.domain_combo.currentText(), w.ALL_DOMAINS_LABEL)
        self.assertFalse(w.domain_combo.isEnabled())

        w.filterLineEdit.setText("da")
        self.assertEqual(self.__titles(w), {"Bax data set"})
        self.assertEqual(w.language_combo.currentText(), "Slovenščina")
        self.assertTrue(w.language_combo.isEnabled())
        self.assertEqual(w.domain_combo.currentText(), w.GENERAL_DOMAIN_LABEL)
        self.assertTrue(w.domain_combo.isEnabled())


        w.filterLineEdit.setText("bar d")
        self.assertEqual(self.__titles(w), {"Bar data set"})

        w.filterLineEdit.setText("bax d")
        self.assertEqual(self.__titles(w), {"Bax data set"})

        w.language_combo.setCurrentText("English")
        w.language_combo.activated.emit(2)
        self.assertEqual(self.__titles(w), {"Bax data set"})

        settings = w.settingsHandler.pack_data(w)

        w2 = self.create_widget(OWDataSets, stored_settings=settings)
        self.wait_until_stop_blocking(w2)
        self.assertEqual(w2.language_combo.currentText(), "English")
        self.assertEqual(self.__titles(w2), {"Foo data set"})

        w.selected_id = "bax.tab"
        settings = w.settingsHandler.pack_data(w)
        w2 = self.create_widget(OWDataSets, stored_settings=settings)
        self.wait_until_stop_blocking(w2)
        self.assertEqual(w2.language_combo.currentText(), w2.ALL_LANGUAGES)
        self.assertFalse(w2.language_combo.isEnabled())
        self.assertEqual(w2.filterLineEdit.text(), "bax d")
        self.assertEqual(self.__titles(w2), {"Bax data set"})


    @patch("Orange.widgets.data.owdatasets.list_remote",
           Mock(return_value={('core', 'foo.tab'): {"language": "English"},
                              ('core', 'bar.tab'): {"language": "Slovenščina"}}))
    @patch("Orange.widgets.data.owdatasets.list_local",
           Mock(return_value={}))
    def test_remember_language(self):
        w = self.create_widget(OWDataSets)  # type: OWDataSets
        self.wait_until_stop_blocking(w)
        w.language_combo.setCurrentText("Slovenščina")
        w.language_combo.activated.emit(w.language_combo.currentIndex())
        settings = w.settingsHandler.pack_data(w)

        w2 = self.create_widget(OWDataSets, stored_settings=settings)
        self.wait_until_stop_blocking(w2)
        self.assertEqual(w2.language_combo.currentText(), "Slovenščina")

        settings["language"] = "Klingon"
        w2 = self.create_widget(OWDataSets, stored_settings=settings)
        self.wait_until_stop_blocking(w2)
        self.assertEqual(w2.language_combo.currentText(), "Klingon")

    @patch("Orange.widgets.data.owdatasets.list_remote",
           Mock(return_value={('core', 'foo.tab'): {"language": "English"},
                              ('core', 'bar.tab'): {"language": "Slovenščina"}}))
    @patch("Orange.widgets.data.owdatasets.list_local",
           Mock(return_value={}))
    def test_remember_all_languages(self):
        w = self.create_widget(OWDataSets)  # type: OWDataSets
        self.wait_until_stop_blocking(w)
        w.language_combo.setCurrentText(w.ALL_LANGUAGES)
        w.language_combo.activated.emit(w.language_combo.currentIndex())
        settings = w.settingsHandler.pack_data(w)

        w2 = self.create_widget(OWDataSets, stored_settings=settings)
        self.wait_until_stop_blocking(w2)
        self.assertEqual(w2.language_combo.currentText(), w2.ALL_LANGUAGES)

    @patch("Orange.widgets.data.owdatasets.list_remote",
           Mock(return_value={('core', 'iris.tab'): {}}))
    @patch("Orange.widgets.data.owdatasets.list_local",
           Mock(return_value={}))
    @patch("Orange.widgets.data.owdatasets.ensure_local",
           Mock(return_value="iris.tab"))
    @WidgetTest.skipNonEnglish
    def test_download_iris(self):
        w = self.create_widget(OWDataSets)  # type: OWDataSets
        self.wait_until_stop_blocking(w)
        # select the only dataset
        sel_type = QItemSelectionModel.ClearAndSelect | QItemSelectionModel.Rows
        w.view.selectionModel().select(w.view.model().index(0, 0), sel_type)
        w.commit()
        self.assertEqual(w.selected_id, "iris.tab")
        iris = self.get_output(w.Outputs.data, w)
        self.assertEqual(len(iris), 150)

    @patch("Orange.widgets.data.owdatasets.list_remote",
           Mock(side_effect=requests.exceptions.ConnectionError))
    @patch("Orange.widgets.data.owdatasets.list_local",
           Mock(return_value={('dir1', 'dir2', 'foo.tab'): {},
                              ('bar.tab',): {}}))
    @patch("Orange.widgets.data.owdatasets.log", Mock())
    @WidgetTest.skipNonEnglish
    def test_dir_depth(self):
        w = self.create_widget(OWDataSets)  # type: OWDataSets
        self.wait_until_stop_blocking(w)
        self.assertEqual(w.view.model().rowCount(), 2)

    def test_migrate_selected_id(self):
        settings = {}
        OWDataSets.migrate_settings(settings, 0)
        self.assertNotIn("selected_id", settings)

        settings = {"selected_id": None}
        OWDataSets.migrate_settings(settings, 0)
        self.assertEqual(settings["selected_id"], None)

        settings = {"selected_id": "dir1\\bar"}
        OWDataSets.migrate_settings(settings, 0)
        self.assertEqual(settings["selected_id"], "bar")

        settings = {"selected_id": "dir1/bar"}
        OWDataSets.migrate_settings(settings, 0)
        self.assertEqual(settings["selected_id"], "bar")

        settings = {"selected_id": "bar"}
        OWDataSets.migrate_settings(settings, 0)
        self.assertEqual(settings["selected_id"], "bar")


if __name__ == "__main__":
    unittest.main()



import numpy as np
import sys
# import warnings
# warnings.filterwarnings("ignore")

from Orange.classification.utils.fasterrisk.utils import get_support_indices, compute_logisticLoss_from_betas_and_yX, insertIntercept_asFirstColOf_X

class starRaySearchModel:
    def __init__(self, X, y, lb=-5, ub=5, num_ray_search=20, early_stop_tolerance=0.001):
        self.X = insertIntercept_asFirstColOf_X(X)
        self.y = y.reshape(-1)
        self.yX = self.y.reshape(-1, 1) * self.X

        self.n = self.X.shape[0]
        self.p = self.X.shape[1]

        if isinstance(ub, (float, int)):
            self.ub_arr = ub * np.ones((self.p, ))
            self.ub_arr[0] = 100.0 # intercept upper bound
        else:
            self.ub_arr = np.insert(ub, 0, 100.0)   # add intercept upper bound
        
        if isinstance(lb, (float, int)):
            self.lb_arr = lb * np.ones((self.p, ))
            self.lb_arr[0] = -100.0 # intercept lower bound
        else:
            self.lb_arr = np.insert(lb, 0, -100)    # add intercept lower bound

        self.num_ray_search = num_ray_search
        self.early_stop_tolerance = early_stop_tolerance
    
    def get_multipliers_for_line_search(self, betas):
        """Get an array of multipliers to try for line search

        Parameters
        ----------
        betas : ndarray
            (1D array with `float` type) a given solution with shape = (1+p, ) assuming the first entry is the intercept

        Returns
        -------
        multipliers : ndarray
            (1D array with `float` type) an array of candidate multipliers with shape = (num_ray_search, )
        """
        # largest_multiplier = min(self.abs_coef_ub/np.max(np.abs(betas[1:])), self.abs_intercept_ub/abs(betas[0]))
        pos_nonzeroIndices = np.where(betas > 1e-8)[0]
        neg_nonzeroIndices = np.where(betas < -1e-8)[0]
        len_pos_nonzeroIndices = len(pos_nonzeroIndices)
        len_neg_nonzeroIndices = len(neg_nonzeroIndices)

        assert len_pos_nonzeroIndices + len_neg_nonzeroIndices > 0, "betas needs to have at least one nonzero entries!"
        largest_multiplier = 1e8
        if len_pos_nonzeroIndices > 0:
            largest_multiplier = min(largest_multiplier, min(self.ub_arr[pos_nonzeroIndices] / betas[pos_nonzeroIndices]))
        if len_neg_nonzeroIndices > 0:
            largest_multiplier = min(largest_multiplier, min(self.lb_arr[neg_nonzeroIndices] / betas[neg_nonzeroIndices]))
        
        if largest_multiplier > 1:
            multipliers = np.linspace(1, largest_multiplier, self.num_ray_search)
        else:
            multipliers = np.linspace(1, 0.5, self.num_ray_search)
        return multipliers
    
    def star_ray_search_scale_and_round(self, sparseDiversePool_beta0_continuous, sparseDiversePool_betas_continuous):
        """For each continuous solution in the sparse diverse pool, find the best multiplier and integer solution. Return the best integer solutions and the corresponding multipliers in the sparse diverse pool

        Parameters
        ----------
        sparseDiversePool_beta_continuous : ndarray
            (1D array with `float` type) an array of continuous intercept with shape = (m, )
        sparseDiversePool_betas_continuous : ndarray
            (2D array with `float` type) an array of continuous coefficients with shape = (m, p) 

        Returns
        -------
        multipliers : ndarray
            (1D array with `float` type) best multiplier for each continuous solution with shape = (m, )
        best_beta0 : ndarray
            (1D array with `float` type) best integer intercept for each continuous solution with shape = (m, )
        best_betas : ndarray
            (2D array with `float` type) best integer coefficient for each continuous solution with shape = (m, p)
        """
        sparseDiversePool_continuous = np.hstack((sparseDiversePool_beta0_continuous.reshape(-1, 1), sparseDiversePool_betas_continuous))
        
        sparseDiversePool_integer = np.zeros(sparseDiversePool_continuous.shape)
        multipliers = np.zeros((sparseDiversePool_integer.shape[0]))

        for i in range(len(multipliers)):
            multipliers[i], sparseDiversePool_integer[i] = self.line_search_scale_and_round(sparseDiversePool_continuous[i])
        
        return multipliers, sparseDiversePool_integer[:, 0], sparseDiversePool_integer[:, 1:]

    def line_search_scale_and_round(self, betas):
        """For a given solution betas, multiply the solution with different multipliers and round each scaled solution to integers. Return the best integer solution based on the logistic loss.

        Parameters
        ----------
        betas : ndarray
            (1D array with `float` type) a given solution with shape = (1+p, ) assuming the first entry is the intercept

        Returns
        -------
        best_multiplier : float
            best multiplier among all pairs of (multiplier, integer_solution)
        best_betas : ndarray
            (1D array with `float` type) best integer solution among all pairs of (multiplier, integer_solution)
        """
        nonzero_indices = get_support_indices(betas)
        num_nonzero = len(nonzero_indices)

        # X_sub = self.X[:, nonzero_indices]
        yX_sub = self.yX[:, nonzero_indices]
        betas_sub = betas[nonzero_indices]

        multipliers = self.get_multipliers_for_line_search(betas_sub)

        loss_continuous_betas = compute_logisticLoss_from_betas_and_yX(betas_sub, yX_sub)
        
        best_multiplier = 1.0
        best_loss = 1e12
        best_betas_sub = np.zeros((num_nonzero, ))

        for multiplier in multipliers:
            betas_sub_scaled = betas_sub * multiplier
            yX_sub_scaled = yX_sub / multiplier

            betas_sub_scaled = self.auxilliary_rounding(betas_sub_scaled, yX_sub_scaled)

            tmp_loss = compute_logisticLoss_from_betas_and_yX(betas_sub_scaled / multiplier, yX_sub)

            if tmp_loss < best_loss:
                best_loss = tmp_loss
                best_multiplier = multiplier
                best_betas_sub[:] = betas_sub_scaled[:]

            if (tmp_loss - loss_continuous_betas) / loss_continuous_betas < self.early_stop_tolerance:
                break

        best_betas = np.zeros((self.p, ))
        best_betas[nonzero_indices] = best_betas_sub

        return best_multiplier, best_betas

    def get_rounding_distance_and_dimension(self, betas):
        """For each dimension, get distances from the real coefficient to the rounded-up integer and the rounded-down integer

        Parameters
        ----------
        betas : ndarray
            (1D array with `float` type) current continuous (real-valued) solution

        Returns
        -------
        betas_floor : ndarray
            (1D array with `float` type) rounded-down coefficients
        dist_from_start_to_floor: ndarray
            (1D array with `float` type) distance from the real coefficient to the rounded-down integer
        betas_ceil : ndarray 
            (1D array with `float` type) rounded-up coefficients
        dist_from_start_to_ceil: ndarray 
            (1D array with `float` type) distance from the real coefficient to the rounded-up integer
        dimensions_to_round: int[:]
            array of indices where the coefficients are not integers to begin with and upon which we should do rounding
        """
        betas_floor = np.floor(betas)
        # floor_is_zero = np.equal(betas_floor, 0)
        dist_from_start_to_floor = betas_floor - betas

        betas_ceil = np.ceil(betas)
        # ceil_is_zero = np.equal(betas_ceil, 0)
        dist_from_start_to_ceil = betas_ceil - betas

        dimensions_to_round = np.flatnonzero(np.not_equal(betas_floor, betas_ceil)).tolist()

        return betas_floor, dist_from_start_to_floor, betas_ceil, dist_from_start_to_ceil, dimensions_to_round

    def auxilliary_rounding(self, betas, yX):
        """Round the solutions to intgers according to the auxilliary loss proposed in the paper

        Parameters
        ----------
        betas : ndarray
            (1D array with `float` type) current continuous (real-valued) solution
        yX : ndarray
            (2D array with `float` type) yX[i, j] = y[i] * X[i, j]

        Returns
        -------
        integer_beta : ndarray
            (1D array with `float` type) rounded integer solution
        """
        n_local, p_local = yX.shape[0], yX.shape[1]

        betas_floor, dist_from_start_to_floor, betas_ceil, dist_from_start_to_ceil, dimensions_to_round = self.get_rounding_distance_and_dimension(betas)

        # yXB = yX.dot(betas) # shape is (n_local, )

        Gamma = np.zeros((n_local, p_local))
        Gamma[:] = betas_floor
        Gamma = Gamma + 1.0 * (yX <= 0)

        yX_Gamma = yX * Gamma
        yXB_extreme = np.sum(yX_Gamma, axis=1)
        l_factors = np.reciprocal((1 + np.exp(yXB_extreme))) # corresponding to l_i's in the NeurIPS paper

        lyX = l_factors.reshape(-1, 1) * yX
        lyX_norm_square = np.sum(lyX * lyX, axis = 0)

        upperBound_arr = 1e12 * np.ones((2 * p_local))
        lyXB_diff = np.zeros((n_local, )) # at the start, betas are not rounded, so coefficient difference is zero
        current_upperBound = 0 # at the start, upper is also 0 because betas have not been rounded yet

        while len(dimensions_to_round) > 0:
            upperBound_arr.fill(1e12)

            for j in dimensions_to_round:
                upperBound_expectation = current_upperBound - lyX_norm_square[j] * dist_from_start_to_floor[j] * dist_from_start_to_ceil[j]

                lyX_j = lyX[:, j]
                lyXB_diff_floor_j = lyXB_diff + dist_from_start_to_ceil[j] * lyX_j
                upperBound_arr[2*j+1] = np.sum(lyXB_diff_floor_j ** 2) # odd positions stores upper bound for ceiling operation

                if upperBound_arr[2*j+1] > upperBound_expectation:
                    lyXB_diff_ceil_j = lyXB_diff + dist_from_start_to_floor[j] * lyX_j
                    upperBound_arr[2*j] = np.sum(lyXB_diff_ceil_j ** 2) # even positions stores upper bound for flooring operation
            
            best_idx_upperBound_arr = np.argmin(upperBound_arr)
            current_upperBound = upperBound_arr[best_idx_upperBound_arr]

            best_j, is_ceil = best_idx_upperBound_arr // 2, best_idx_upperBound_arr % 2

            if is_ceil:
                betas[best_j] += dist_from_start_to_ceil[best_j]
                lyXB_diff = lyXB_diff + dist_from_start_to_ceil[best_j] * lyX[:, best_j]
            else:
                betas[best_j] += dist_from_start_to_floor[best_j]
                lyXB_diff = lyXB_diff + dist_from_start_to_floor[best_j] * lyX[:, best_j]
            
            dimensions_to_round.remove(best_j)
        
        return betas


from copy import copy
import json
from os import path, makedirs

from AnyQt.QtGui import QImage, QPainter
from AnyQt.QtWidgets import QGraphicsScene, QApplication, QWidget, QGraphicsView, QHBoxLayout
from AnyQt.QtCore import QRectF, Qt, QTimer
# QWebEngineWidgets must be imported before QCoreApplication is created.
# It will fail with an import error if imported (first time) after.
import AnyQt.QtWebEngineWidgets  # pylint: disable=unused-import

from orangecanvas.canvas.items import NodeItem
from orangecanvas.help import HelpManager
from orangecanvas.registry import WidgetRegistry

from Orange.canvas.config import Config as OConfig


class WidgetCatalog:
    def __init__(self, output_dir, image_url_prefix):
        self.output_dir = output_dir
        self.image_url_prefix = image_url_prefix

        QApplication.setAttribute(Qt.AA_EnableHighDpiScaling)
        QApplication.setAttribute(Qt.AA_UseHighDpiPixmaps)
        QApplication.setAttribute(Qt.AA_ShareOpenGLContexts)
        self.app = QApplication([])

        print("Generating widget repository")
        self.registry = self.__get_widget_registry()
        print("Locating help files")
        self.help_manager = HelpManager()
        self.help_manager.set_registry(self.registry)

        # Help manager needs QApplication running to get the urls
        # of widget documentation. 5 seconds should be enough.
        QTimer.singleShot(5000, self.app.quit)
        self.app.exec()

        self.__scene = QGraphicsScene()
        self.__nodes = []
        print("Ready to go")

    def create(self):
        print("Generating catalog")
        try:
            makedirs(path.join(self.output_dir, "icons"))
        except FileExistsError:
            pass

        result = []
        for category in self.registry.categories():
            widgets = []
            result.append((category.name, widgets))
            for widget in category.widgets:
                widgets.append({
                    "text": widget.name,
                    "doc": self.__get_help(widget),
                    "img": self.__get_icon(widget, category),
                    "keyword": widget.keywords,
                })

        with open(path.join(self.output_dir, "widgets.json"), 'wt') as f:
            json.dump(result, f, indent=1)
        print("Done")

    @staticmethod
    def __get_widget_registry():
        widget_registry = WidgetRegistry()
        widget_discovery = OConfig.widget_discovery(widget_registry)
        widget_discovery.run(OConfig.widgets_entry_points())

        # Fixup category.widgets list
        for cat, widgets in widget_registry._categories_dict.values():
            cat.widgets = widgets
        return widget_registry

    def __get_icon(self, widget, category=None):
        # Set remove inputs/outputs so the "ears" are not drawn
        widget = copy(widget)
        widget.inputs = []
        widget.outputs = []

        w = IconWidget()
        w.set_widget(widget, category)
        w.show()
        #self.app.processEvents()
        filename = "icons/{}.png".format(widget.qualified_name)
        w.render_as_png(path.join(self.output_dir, filename))
        w.hide()

        if self.image_url_prefix:
            return self.image_url_prefix + filename
        else:
            return filename

    def __get_help(self, widget):
        query = dict(id=widget.qualified_name)
        try:
            return self.help_manager.search(query).url()
        except KeyError:
            return None


class IconWidget(QWidget):
    def __init__(self):
        super().__init__()
        self.setLayout(QHBoxLayout())
        self.layout().setContentsMargins(0, 0, 0, 0)
        self.setFixedSize(50, 50)

        view = QGraphicsView()
        self.layout().addWidget(view)
        self.scene = QGraphicsScene(view)
        view.setScene(self.scene)

    def set_widget(self, widget_description, category_description):
        node = NodeItem(widget_description)
        if category_description is not None:
            node.setWidgetCategory(category_description)
        self.scene.addItem(node)

    def render_as_png(self, filename):
        img = QImage(50, 50, QImage.Format_ARGB32)
        img.fill(Qt.transparent)
        painter = QPainter(img)
        painter.setRenderHint(QPainter.Antialiasing, 1)
        self.scene.render(painter, QRectF(0, 0, 50, 50), QRectF(-25, -25, 50, 50))
        painter.end()
        img.save(filename)


if __name__ == '__main__':
    from optparse import OptionParser

    parser = OptionParser(usage="usage: %prog --output <outputdir> [options]")
    parser.add_option('--url-prefix', dest="prefix",
                      help="prefix to prepend to image urls")
    parser.add_option('--output', dest="output",
                      help="path where widgets.json will be created")

    options, args = parser.parse_args()
    if not options.output:
        parser.error("Please specify the output dir")

    w = WidgetCatalog(output_dir=options.output, image_url_prefix=options.prefix)
    w.create()



import numpy

import Orange.data
from Orange.widgets.widget import OWWidget, Input, Output, settings
from Orange.widgets.utils.widgetpreview import WidgetPreview
from Orange.widgets import gui


class OWDataSamplerC(OWWidget):
    name = "Data Sampler (C)"
    description = "Randomly selects a subset of instances from the dataset."
    icon = "icons/DataSamplerC.svg"
    priority = 30

    class Inputs:
        data = Input("Data", Orange.data.Table)

    # [start-snippet-1]
    class Outputs:
        sample = Output("Sampled Data", Orange.data.Table)
        other = Output("Other Data", Orange.data.Table)

    # [end-snippet-1]
    proportion = settings.Setting(50)
    commitOnChange = settings.Setting(0)

    want_main_area = False

    def __init__(self):
        super().__init__()

        self.dataset = None
        self.sample = None
        self.otherdata = None

        # GUI
        box = gui.widgetBox(self.controlArea, "Info")
        self.infoa = gui.widgetLabel(
            box, "No data on input yet, waiting to get something."
        )
        self.infob = gui.widgetLabel(box, "")

        gui.separator(self.controlArea)
        self.optionsBox = gui.widgetBox(self.controlArea, "Options")
        gui.spin(
            self.optionsBox,
            self,
            "proportion",
            minv=10,
            maxv=90,
            step=10,
            label="Sample Size [%]:",
            callback=[self.selection, self.checkCommit],
        )
        gui.checkBox(
            self.optionsBox, self, "commitOnChange", "Commit data on selection change"
        )
        gui.button(self.optionsBox, self, "Commit", callback=self.commit)
        self.optionsBox.setDisabled(True)

        self.resize(100, 50)

    @Inputs.data
    def set_data(self, dataset):
        if dataset is not None:
            self.dataset = dataset
            self.infoa.setText("%d instances in input dataset" % len(dataset))
            self.optionsBox.setDisabled(False)
            self.selection()
        else:
            self.sample = None
            self.otherdata = None
            self.optionsBox.setDisabled(True)
            self.infoa.setText("No data on input yet, waiting to get something.")
            self.infob.setText("")
        self.commit()

    def selection(self):
        if self.dataset is None:
            return

        n_selected = int(numpy.ceil(len(self.dataset) * self.proportion / 100.0))
        indices = numpy.random.permutation(len(self.dataset))
        indices_sample = indices[:n_selected]
        indices_other = indices[n_selected:]
        self.sample = self.dataset[indices_sample]
        self.otherdata = self.dataset[indices_other]
        self.infob.setText("%d sampled instances" % len(self.sample))

    def commit(self):
        self.Outputs.sample.send(self.sample)
        self.Outputs.sample.send(self.otherdata)

    def checkCommit(self):
        if self.commitOnChange:
            self.commit()


if __name__ == "__main__":
    WidgetPreview(OWDataSamplerC).run(Orange.data.Table("iris"))



import numpy as np
import sys
# import warnings
# warnings.filterwarnings("ignore")

from Orange.classification.utils.fasterrisk.utils import get_support_indices, get_nonsupport_indices, compute_logisticLoss_from_ExpyXB
from Orange.classification.utils.fasterrisk.base_model import logRegModel

class sparseLogRegModel(logRegModel):
    def __init__(self, X, y, lambda2=1e-8, intercept=True, original_lb=-5, original_ub=5):
        super().__init__(X=X, y=y, lambda2=lambda2, intercept=intercept, original_lb=original_lb, original_ub=original_ub)
    
    def getAvailableIndices_for_expansion(self, betas):
        """Get the indices of features that can be added to the support of the current sparse solution

        Parameters
        ----------
        betas : ndarray
            (1D array with `float` type) The current sparse solution

        Returns
        -------
        available_indices : ndarray
            (1D array with `int` type) The indices of features that can be added to the support of the current sparse solution
        """
        available_indices = get_nonsupport_indices(betas)
        return available_indices
   
    def expand_parent_i_support_via_OMP_by_1(self, i, child_size=10):
        """For parent solution i, generate [child_size] child solutions

        Parameters
        ----------
        i : int
            index of the parent solution
        child_size : int, optional
            how many child solutions to generate based on parent solution i, by default 10
        """
        # non_support = get_nonsupport_indices(self.betas_arr_parent[i])
        non_support = self.getAvailableIndices_for_expansion(self.betas_arr_parent[i])
        support = get_support_indices(self.betas_arr_parent[i])

        grad_on_non_support = self.yXT[non_support].dot(np.reciprocal(1+self.ExpyXB_arr_parent[i]))
        abs_grad_on_non_support = np.abs(grad_on_non_support)

        num_new_js = min(child_size, len(non_support))
        new_js = non_support[np.argsort(-abs_grad_on_non_support)][:num_new_js]
        child_start, child_end = i*child_size, i*child_size + num_new_js

        self.ExpyXB_arr_child[child_start:child_end] = self.ExpyXB_arr_parent[i, :] # (num_new_js, n)
        # self.betas_arr_child[child_start:child_end, non_support] = 0
        self.betas_arr_child[child_start:child_end] = 0
        self.betas_arr_child[child_start:child_end, support] = self.betas_arr_parent[i, support]
        self.beta0_arr_child[child_start:child_end] = self.beta0_arr_parent[i]
        
        beta_new_js = np.zeros((num_new_js, )) #(len(new_js), )
        diff_max = 1e3

        step = 0
        while step < 10 and diff_max > 1e-3:
            prev_beta_new_js = beta_new_js.copy()
            grad_on_new_js = -np.sum(self.yXT[new_js] * np.reciprocal(1.+self.ExpyXB_arr_child[child_start:child_end]), axis=1) + self.twoLambda2 * beta_new_js
            step_at_new_js = grad_on_new_js / self.Lipschitz

            beta_new_js = prev_beta_new_js - step_at_new_js
            beta_new_js = np.clip(beta_new_js, self.lbs[new_js], self.ubs[new_js])
            diff_beta_new_js = beta_new_js - prev_beta_new_js

            self.ExpyXB_arr_child[child_start:child_end] *= np.exp(self.yXT[new_js] * diff_beta_new_js.reshape(-1, 1))

            diff_max = max(np.abs(diff_beta_new_js))
            step += 1

        for l in range(num_new_js):
            child_id = child_start + l
            self.betas_arr_child[child_id, new_js[l]] = beta_new_js[l]
            tmp_support_str = str(get_support_indices(self.betas_arr_child[child_id]))
            if tmp_support_str not in self.forbidden_support:
                self.total_child_added += 1 # count how many unique child has been added for a specified support size
                self.forbidden_support.add(tmp_support_str)

                self.ExpyXB_arr_child[child_id], self.beta0_arr_child[child_id], self.betas_arr_child[child_id] = self.finetune_on_current_support(self.ExpyXB_arr_child[child_id], self.beta0_arr_child[child_id], self.betas_arr_child[child_id])
                self.loss_arr_child[child_id] = compute_logisticLoss_from_ExpyXB(self.ExpyXB_arr_child[child_id])

    def beamSearch_multipleSupports_via_OMP_by_1(self, parent_size=10, child_size=10):
        """Each parent solution generates [child_size] child solutions, so there will be [parent_size] * [child_size] number of total child solutions. However, only the top [parent_size] child solutions are retained as parent solutions for the next level i+1.

        Parameters
        ----------
        parent_size : int, optional
            how many top solutions to retain at each level, by default 10
        child_size : int, optional
            how many child solutions to generate based on each parent solution, by default 10
        """
        self.loss_arr_child.fill(1e12)
        self.total_child_added = 0

        for i in range(self.num_parent):
            self.expand_parent_i_support_via_OMP_by_1(i, child_size=child_size)

        child_indices = np.argsort(self.loss_arr_child)[:min(parent_size, self.total_child_added)] # get indices of children which have the smallest losses
        num_child_indices = len(child_indices)
        self.ExpyXB_arr_parent[:num_child_indices], self.beta0_arr_parent[:num_child_indices], self.betas_arr_parent[:num_child_indices] = self.ExpyXB_arr_child[child_indices], self.beta0_arr_child[child_indices], self.betas_arr_child[child_indices]

        self.num_parent = num_child_indices

    def get_sparse_sol_via_OMP(self, k, parent_size=10, child_size=10):
        """Get sparse solution through beam search and orthogonal matching pursuit (OMP), for level i, each parent solution generates [child_size] child solutions, so there will be [parent_size] * [child_size] number of total child solutions. However, only the top [parent_size] child solutions are retained as parent solutions for the next level i+1.

        Parameters
        ----------
        k : int
            number of nonzero coefficients for the final sparse solution
        parent_size : int, optional
            how many top solutions to retain at each level, by default 10
        child_size : int, optional
            how many child solutions to generate based on each parent solution, by default 10
        """
        nonzero_indices_set = set(np.where(np.abs(self.betas) > 1e-9)[0])
        # print("get_sparse_sol_via_OMP, initial support is:", nonzero_indices_set)
        zero_indices_set = set(range(self.p)) - nonzero_indices_set
        num_nonzero = len(nonzero_indices_set)

        if len(zero_indices_set) == 0:
            return

        # if there is no warm start solution, initialize beta0 analytically
        if (self.intercept) and (len(nonzero_indices_set) == 0):
            y_sum = np.sum(self.y)
            num_y_pos_1 = (y_sum + self.n)/2
            num_y_neg_1 = self.n - num_y_pos_1
            self.beta0 = np.log(num_y_pos_1/num_y_neg_1)
            self.ExpyXB *= np.exp(self.y * self.beta0)

        # create beam search parent
        self.ExpyXB_arr_parent = np.zeros((parent_size, self.n))
        self.beta0_arr_parent = np.zeros((parent_size, ))
        self.betas_arr_parent = np.zeros((parent_size, self.p))
        self.ExpyXB_arr_parent[0, :] = self.ExpyXB[:]
        self.beta0_arr_parent[0] = self.beta0
        self.betas_arr_parent[0, :] = self.betas[:]
        self.num_parent = 1

        # create beam search children. parent[i]->child[i*child_size:(i+1)*child_size]
        total_child_size = parent_size * child_size
        self.ExpyXB_arr_child = np.zeros((total_child_size, self.n))
        self.beta0_arr_child = np.zeros((total_child_size, ))
        self.betas_arr_child = np.zeros((total_child_size, self.p))
        self.isMasked_arr_child = np.ones((total_child_size, ), dtype=bool)
        self.loss_arr_child = 1e12 * np.ones((total_child_size, ))
        self.forbidden_support = set()

        while num_nonzero < min(k, self.p):
            num_nonzero += 1
            self.beamSearch_multipleSupports_via_OMP_by_1(parent_size=parent_size, child_size=child_size)

        self.ExpyXB, self.beta0, self.betas = self.ExpyXB_arr_parent[0], self.beta0_arr_parent[0], self.betas_arr_parent[0]

class groupSparseLogRegModel(sparseLogRegModel):
    def __init__(self, X, y, lambda2=1e-8, intercept=True, original_lb=-5, original_ub=5, group_sparsity=10, featureIndex_to_groupIndex=None, groupIndex_to_featureIndices=None):
        super().__init__(X=X, y=y, lambda2=lambda2, intercept=intercept, original_lb=original_lb, original_ub=original_ub)

        self.group_sparsity = group_sparsity
        self.featureIndex_to_groupIndex = featureIndex_to_groupIndex # this is a numpy array
        self.groupIndex_to_featureIndices = groupIndex_to_featureIndices # this is a dictionary of sets
    
    def getAvailableIndices_for_expansion(self, betas):
        """Get the indices of features that can be added to the support of the current sparse solution

        Parameters
        ----------
        betas : ndarray
            (1D array with `float` type) The current sparse solution

        Returns
        -------
        available_indices : ndarray
            (1D array with `int` type) The indices of features that can be added to the support of the current sparse solution
        """
        support = get_support_indices(betas)
        existing_groupIndices = np.unique(self.featureIndex_to_groupIndex[support])
        if len(existing_groupIndices) < self.group_sparsity:
            available_indices = get_nonsupport_indices(betas)
        else:
            available_indices = set()
            for groupIndex in existing_groupIndices:
                available_indices.update(self.groupIndex_to_featureIndices[groupIndex])
            available_indices = available_indices - set(support)
            available_indices = np.array(list(available_indices), dtype=int)

        return available_indices
   


import numpy as np
import sklearn.metrics

from Orange.classification.utils.fasterrisk.sparseBeamSearch import sparseLogRegModel, groupSparseLogRegModel
from Orange.classification.utils.fasterrisk.sparseDiversePool import sparseDiversePoolLogRegModel, groupSparseDiversePoolLogRegModel
from Orange.classification.utils.fasterrisk.rounding import starRaySearchModel

from Orange.classification.utils.fasterrisk.utils import compute_logisticLoss_from_X_y_beta0_betas, get_all_product_booleans, get_support_indices, get_all_product_booleans, get_groupIndex_to_featureIndices, check_bounds

class RiskScoreOptimizer:
    def __init__(self, X, y, k, select_top_m=50, lb=-5, ub=5, \
                 gap_tolerance=0.05, parent_size=10, child_size=None, \
                 maxAttempts=50, num_ray_search=20, \
                 lineSearch_early_stop_tolerance=0.001, \
                 group_sparsity=None, featureIndex_to_groupIndex=None):
        """Initialize the RiskScoreOptimizer class, which performs sparseBeamSearch and generates integer sparseDiverseSet

        Parameters
        ----------
        X : ndarray
            (2D array with `float` type) feature matrix, each row[i, :] corresponds to the features of sample i
        y : ndarray
            (1D array with `float` type) labels (+1 or -1) of each sample
        k : int
            number of selected features in the final sparse model
        select_top_m : int, optional
            number of top solutions to keep among the pool of diverse sparse solutions, by default 50
        lb : float or list, optional
            lower bound(s) of the coefficients, when passed as a list, specifies lower bounds for all the features in X, by default -5
        ub : float or list, optional
            upper bound(s) of the coefficients, when passed as a list, specifies lower bounds for all the features in X, by default 5
        parent_size : int, optional
            how many solutions to retain after beam search, by default 10
        child_size : int, optional
            how many new solutions to expand for each existing solution, by default None
        maxAttempts : int, optional
            how many alternative features to try in order to replace the old feature during the diverse set pool generation, by default None
        num_ray_search : int, optional
            how many multipliers to try for each continuous sparse solution, by default 20
        lineSearch_early_stop_tolerance : float, optional
            tolerance level to stop linesearch early (error_of_loss_difference/loss_of_continuous_solution), by default 0.001
        group_sparsity : int, optional
            number of groups to be selected, by default None
        featureIndex_to_groupIndex : ndarray, optional
            (1D array with `int` type) featureIndex_to_groupIndex[i] is the group index of feature i, by default None
        """

        # check the formats of inputs X and y
        y_shape = y.shape
        y_unique = np.unique(y)
        y_unique_expected = np.asarray([-1, 1])
        X_shape = X.shape
        assert len(y_shape) == 1, "input y must have 1-D shape!"
        assert len(y_unique) == 2, "input y must have only 2 labels"
        assert max(np.abs(y_unique - y_unique_expected)) < 1e-8, "input y must be equal to only +1 or -1"
        assert len(X_shape) == 2, "input X must have 2-D shape!"
        assert X_shape[0] == y_shape[0], "number of rows from input X must be equal to the number of elements from input y!"
        self.y = y
        self.X = X

        self.k = k
        self.parent_size = parent_size
        self.child_size = self.parent_size
        if child_size is not None:
            self.child_size = child_size

        self.sparseDiverseSet_gap_tolerance = gap_tolerance
        self.sparseDiverseSet_select_top_m = select_top_m
        self.sparseDiverseSet_maxAttempts = maxAttempts

        lb = check_bounds(lb, 'lb', X_shape[1])
        ub = check_bounds(ub, 'ub', X_shape[1])

        self.group_sparsity = group_sparsity
        self.featureIndex_to_groupIndex = featureIndex_to_groupIndex

        if self.group_sparsity is None:
            self.sparseLogRegModel_object = sparseLogRegModel(X, y, intercept=True, original_lb=lb, original_ub=ub)
            self.sparseDiversePoolLogRegModel_object = sparseDiversePoolLogRegModel(X, y, intercept=True, original_lb=lb, original_ub=ub)
        else:
            assert type(group_sparsity) == int, "group_sparsity needs to be an integer"
            assert group_sparsity > 0, "group_sparsity needs to be > 0!"
            assert group_sparsity > 0, "group_sparsity needs to be > 0!"

            assert self.featureIndex_to_groupIndex is not None, "featureIndex_to_groupIndex must be provided if group_sparsity is not None"
            assert type(self.featureIndex_to_groupIndex[0]) == np.int_, "featureIndex_to_groupIndex needs to be a NumPy integer array"
        
            self.groupIndex_to_featureIndices = get_groupIndex_to_featureIndices(self.featureIndex_to_groupIndex)

            self.sparseLogRegModel_object = groupSparseLogRegModel(X, y, intercept=True, original_lb=lb, original_ub=ub, group_sparsity=self.group_sparsity, featureIndex_to_groupIndex=self.featureIndex_to_groupIndex, groupIndex_to_featureIndices=self.groupIndex_to_featureIndices)
            self.sparseDiversePoolLogRegModel_object = groupSparseDiversePoolLogRegModel(X, y, intercept=True, original_lb=lb, original_ub=ub, group_sparsity=self.group_sparsity, featureIndex_to_groupIndex=self.featureIndex_to_groupIndex, groupIndex_to_featureIndices=self.groupIndex_to_featureIndices)

        self.starRaySearchModel_object = starRaySearchModel(X = X, y = y, lb=lb, ub=ub, num_ray_search=num_ray_search, early_stop_tolerance=lineSearch_early_stop_tolerance)

        self.IntegerPoolIsSorted = False

    def optimize(self):
        """performs sparseBeamSearch, generates integer sparseDiverseSet, and perform star ray search
        """
        self.sparseLogRegModel_object.get_sparse_sol_via_OMP(k=self.k, parent_size=self.parent_size, child_size=self.child_size)
        
        beta0, betas, ExpyXB = self.sparseLogRegModel_object.get_beta0_betas_ExpyXB()
        self.sparseDiversePoolLogRegModel_object.warm_start_from_beta0_betas_ExpyXB(beta0 = beta0, betas = betas, ExpyXB = ExpyXB)
        sparseDiversePool_beta0, sparseDiversePool_betas = self.sparseDiversePoolLogRegModel_object.get_sparseDiversePool(gap_tolerance=self.sparseDiverseSet_gap_tolerance, select_top_m=self.sparseDiverseSet_select_top_m, maxAttempts=self.sparseDiverseSet_maxAttempts)

        self.multipliers, self.sparseDiversePool_beta0_integer, self.sparseDiversePool_betas_integer = self.starRaySearchModel_object.star_ray_search_scale_and_round(sparseDiversePool_beta0, sparseDiversePool_betas)

    def _sort_IntegerPool_on_logisticLoss(self):
        """sort the integer solutions in the pool by ascending order of logistic loss
        """
        sparseDiversePool_XB = (self.sparseDiversePool_beta0_integer.reshape(1, -1) + self.X @ self.sparseDiversePool_betas_integer.transpose()) / (self.multipliers.reshape(1, -1))
        sparseDiversePool_yXB = self.y.reshape(-1, 1) * sparseDiversePool_XB
        sparseDiversePool_ExpyXB = np.exp(sparseDiversePool_yXB)
        # print(sparseDiversePool_ExpyXB.shape)
        sparseDiversePool_logisticLoss = np.sum(np.log(1.+np.reciprocal(sparseDiversePool_ExpyXB)), axis=0)
        orderedIndices = np.argsort(sparseDiversePool_logisticLoss)

        self.multipliers = self.multipliers[orderedIndices]
        self.sparseDiversePool_beta0_integer = self.sparseDiversePool_beta0_integer[orderedIndices]
        self.sparseDiversePool_betas_integer = self.sparseDiversePool_betas_integer[orderedIndices]

        self.IntegerPoolIsSorted = True

    def get_models(self, model_index=None):
        """get risk score models

        Parameters
        ----------
        model_index : int, optional
            index of the model in the integer sparseDiverseSet, by default None

        Returns
        -------
        multipliers : ndarray 
            (1D array with `float` type) multipliers with each entry as multipliers[i] 
        sparseDiversePool_integer : ndarray
            (2D array with `float` type) integer coefficients (intercept included) with each row as an integer solution sparseDiversePool_integer[i]
        """
        if self.IntegerPoolIsSorted is False:
            self._sort_IntegerPool_on_logisticLoss()
        if model_index is not None:
            return self.multipliers[model_index], self.sparseDiversePool_beta0_integer[model_index], self.sparseDiversePool_betas_integer[model_index]
        return self.multipliers, self.sparseDiversePool_beta0_integer, self.sparseDiversePool_betas_integer



class RiskScoreClassifier:
    def __init__(self, multiplier, intercept, coefficients, featureNames = None, X_train = None):
        """Initialize a risk score classifier. Then we can use this classifier to predict labels, predict probabilites, and calculate total logistic loss

        Parameters
        ----------
        multiplier : float
            multiplier of the risk score model
        intercept : float
            intercept of the risk score model
        coefficients : ndarray
            (1D array with `float` type) coefficients of the risk score model
        """
        self.multiplier = multiplier
        self.intercept = intercept
        self.coefficients = coefficients

        self.scaled_intercept = self.intercept / self.multiplier
        self.scaled_coefficients = self.coefficients / self.multiplier

        self.X_train = X_train

        self.reset_featureNames(featureNames)

    def predict(self, X):
        """Predict labels

        Parameters
        ----------
        X : ndarray
            (2D array with `float` type) feature matrix with shape (n, p)

        Returns
        -------
        y_pred : ndarray
            (1D array with `float` type) predicted labels (+1.0 or -1.0) with shape (n, ) 
        """
        y_score = (self.intercept + X.dot(self.coefficients)) / self.multiplier # numpy dot.() has some floating point error issues, so we avoid using self.scaled_intercept and self.scaled_coefficients directly
        y_pred = 2 * (y_score > 0) - 1
        return y_pred

    def predict_prob(self, X):
        """Calculate the risk probabilities of predicting each sample y_i with label +1

        Parameters
        ----------
        X : ndarray
            (2D array with `float` type) feature matrix with shape (n, p)

        Returns
        -------
        y_pred_prob : ndarray
            (1D array with `float` type) probabilities of each sample y_i to be +1 with shape (n, )
        """
        y_score = (self.intercept + X.dot(self.coefficients)) / self.multiplier # numpy dot.() has some floating point error issues, so we avoid using self.scaled_intercept and self.scaled_coefficients directly
        y_pred_prob = 1/(1+np.exp(-y_score))

        return y_pred_prob

    def compute_logisticLoss(self, X, y):
        """Compute the total logistic loss given the feature matrix X and labels y

        Parameters
        ----------
        X : ndarray
            (2D array with `float` type) feature matrix with shape (n, p)
        y : ndarray
            (1D array with `float` type) sample labels (+1 or -1) with shape (n)

        Returns
        -------
        logisticLoss: float
            total logistic loss, loss = $sum_{i=1}^n log(1+exp(-y_i * (beta0 + X[i, :] @ beta) / multiplier))$
        """
        return compute_logisticLoss_from_X_y_beta0_betas(X, y, self.scaled_intercept, self.scaled_coefficients)

    def get_acc_and_auc(self, X, y):
        """Calculate ACC and AUC of a certain dataset with features X and label y

        Parameters
        ----------
        X : ndarray
            (2D array with `float` type) 2D array storing the features
        y : ndarray
            (1D array with `float` type) storing the labels (+1/-1) 

        Returns
        -------
        acc: float
            accuracy
        auc: float
            area under the ROC curve
        """
        y_pred = self.predict(X)
        # print(y_pred.shape, y.shape)
        acc = np.sum(y_pred == y) / len(y)
        y_pred_prob = self.predict_prob(X)

        fpr, tpr, thresholds = sklearn.metrics.roc_curve(y_true=y, y_score=y_pred_prob, drop_intermediate=False)
        auc = sklearn.metrics.auc(fpr, tpr)
        return acc, auc

    def reset_featureNames(self, featureNames):
        """Reset the feature names in the class in order to print out the model card for the user

        Parameters
        ----------
        featureNames : str[:]
            a list of strings which are the feature names for columns of X
        """
        self.featureNames = featureNames

    def _print_score_calculation_table(self):
        assert self.featureNames is not None, "please pass the featureNames to the model by using the function .reset_featureNames(featureNames)"

        nonzero_indices = get_support_indices(self.coefficients)

        max_feature_length = max([len(featureName) for featureName in self.featureNames])
        row_score_template = '{0}. {1:>%d}     {2:>2} point(s) | + ...' % (max_feature_length)

        print("The Risk Score is:")
        for count, feature_i in enumerate(nonzero_indices):
            row_score_str = row_score_template.format(count+1, self.featureNames[feature_i], int(self.coefficients[feature_i]))
            if count == 0:
                row_score_str = row_score_str.replace("+", " ")
            
            print(row_score_str)

        final_score_str = ' ' * (14+max_feature_length) + 'SCORE | =    '
        print(final_score_str)

    def _print_score_risk_row(self, scores, risks):
        score_row = "SCORE |"
        risk_row = "RISK  |"
        score_entry_template = '  {0:>4}  |'
        risk_entry_template = ' {0:>5}% |'
        for (score, risk) in zip(scores, risks):
            score_row += score_entry_template.format(score)
            risk_row += risk_entry_template.format(round(100*risk, 1))
        print(score_row)
        print(risk_row)

    def _print_score_risk_table(self, quantile_len):

        nonzero_indices = get_support_indices(self.coefficients)
        len_nonzero_indices = len(nonzero_indices)

        if len_nonzero_indices <= 10:
            ### method 1: get all possible scores; Drawback for large support size, get the product booleans is too many
            all_product_booleans = get_all_product_booleans(len_nonzero_indices)
            all_scores = all_product_booleans.dot(self.coefficients[nonzero_indices])
            all_scores = np.unique(all_scores)
        else:
            # ### method 2: calculate all scores in the training set, pick the top 20 quantile points
            assert self.X_train is not None, "There are more than 10 nonzero coefficients for the risk scoring system. The number of possible total scores is too many!\n\nPlease consider re-initialize your RiskScoreClassifier_m by providing the training dataset features X_train as follows:\n\n    RiskScoreClassifier_m = RiskScoreClassifier(multiplier, intercept, coefficients, X_train = X_train)"

            all_scores = self.X_train.dot(self.coefficients)
            all_scores = np.unique(all_scores)
            quantile_len = min(quantile_len, len(all_scores))
            quantile_points = np.asarray(range(1, 1+quantile_len)) / quantile_len
            all_scores = np.quantile(all_scores, quantile_points, method = "closest_observation")

        all_scaled_scores = (self.intercept + all_scores) / self.multiplier
        all_risks = 1 / (1 + np.exp(-all_scaled_scores))

        num_scores_div_2 = (len(all_scores) + 1) // 2
        self._print_score_risk_row(all_scores[:num_scores_div_2], all_risks[:num_scores_div_2])
        self._print_score_risk_row(all_scores[num_scores_div_2:], all_risks[num_scores_div_2:])

    def print_model_card(self, quantile_len=20):
        """Print the score evaluation table and score risk table onto terminal
        """
        self._print_score_calculation_table()
        self._print_score_risk_table(quantile_len = quantile_len)


from functools import partial

import numpy as np
import scipy.sparse

from Orange.data import Table, ContinuousVariable, Domain
from Orange.tests.test_table import preprocess_domain_single, preprocess_domain_shared

from .base import Benchmark, benchmark


def add_unknown_attribute(table):
    new_domain = Domain(list(table.domain.attributes) + [ContinuousVariable("x")])
    return table.transform(new_domain)


def add_unknown_class(table):
    new_domain = Domain(table.domain.attributes, class_vars=[ContinuousVariable("x")])
    return table.transform(new_domain)


class BenchTransform(Benchmark):

    def setup_dense(self, rows, cols):
        self.table = Table.from_numpy(  # pylint: disable=W0201
            Domain([ContinuousVariable(str(i)) for i in range(cols)]),
            np.random.RandomState(0).rand(rows, cols))

    def setup_sparse(self, rows, cols):
        sparse = scipy.sparse.rand(rows, cols, density=0.01, format='csr', random_state=0)
        self.table = Table.from_numpy(  # pylint: disable=W0201
            Domain([ContinuousVariable(str(i), sparse=True) for i in range(cols)]),
            sparse)

    @benchmark(setup=partial(setup_dense, rows=10000, cols=100), number=5)
    def bench_copy_dense_long(self):
        add_unknown_attribute(self.table)

    @benchmark(setup=partial(setup_dense, rows=1000, cols=1000), number=5)
    def bench_copy_dense_square(self):
        add_unknown_attribute(self.table)

    @benchmark(setup=partial(setup_dense, rows=100, cols=10000), number=2)
    def bench_copy_dense_wide(self):
        add_unknown_attribute(self.table)

    @benchmark(setup=partial(setup_sparse, rows=10000, cols=100), number=5)
    def bench_copy_sparse_long(self):
        t = add_unknown_attribute(self.table)
        self.assertIsInstance(t.X, scipy.sparse.csr_matrix)

    @benchmark(setup=partial(setup_sparse, rows=1000, cols=1000), number=5)
    def bench_copy_sparse_square(self):
        t = add_unknown_attribute(self.table)
        self.assertIsInstance(t.X, scipy.sparse.csr_matrix)

    @benchmark(setup=partial(setup_sparse, rows=100, cols=10000), number=2)
    def bench_copy_sparse_wide(self):
        t = add_unknown_attribute(self.table)
        self.assertIsInstance(t.X, scipy.sparse.csr_matrix)

    @benchmark(setup=partial(setup_dense, rows=10000, cols=100), number=5)
    def bench_subarray_dense_long(self):
        # adding a class should link X
        add_unknown_class(self.table)

    def setup_dense_transforms(self, rows, cols, transforms):
        self.setup_dense(rows, cols)
        self.domains = []  # pylint: disable=attribute-defined-outside-init
        self.callbacks = []  # pylint: disable=attribute-defined-outside-init
        domain = self.table.domain
        for t in transforms:
            if t == "single":
                call_cv = None
                domain = preprocess_domain_single(domain, call_cv)
                self.callbacks.append((call_cv,))
            elif t == "shared":
                call_cv, call_shared = None, None
                domain = preprocess_domain_shared(domain, call_cv, call_shared)
                self.callbacks.append((call_cv, call_shared))
            else:
                raise RuntimeError
            self.domains.append(domain)

    @benchmark(setup=partial(setup_dense_transforms, rows=1000, cols=100,
                             transforms=["single"]), number=5)
    def bench_transform_single(self):
        t = self.table.transform(self.domains[-1])
        np.testing.assert_almost_equal(t.X, self.table.X * 2)

    @benchmark(setup=partial(setup_dense_transforms, rows=1000, cols=100,
                             transforms=["single", "single"]), number=5)
    def bench_transform_single_single(self):
        t = self.table.transform(self.domains[-1])
        np.testing.assert_almost_equal(t.X, self.table.X * 2**2)

    @benchmark(setup=partial(setup_dense_transforms, rows=1000, cols=100,
                             transforms=["shared"]), number=5)
    def bench_transform_shared(self):
        t = self.table.transform(self.domains[-1])
        np.testing.assert_almost_equal(t.X, self.table.X * 2)

    @benchmark(setup=partial(setup_dense_transforms, rows=1000, cols=100,
                             transforms=["single", "single", "shared", "single"]), number=5)
    def bench_transform_single_single_shared_single(self):
        t = self.table.transform(self.domains[-1])
        np.testing.assert_almost_equal(t.X, self.table.X * 2**4)


    @benchmark(setup=partial(setup_dense_transforms, rows=1000, cols=100,
                             transforms=["single", "single", "shared",
                                         "single", "shared", "single"]), number=5)
    def bench_transform_single_single_shared_single_shared_single(self):
        t = self.table.transform(self.domains[-1])
        np.testing.assert_almost_equal(t.X, self.table.X * 2**6)



from abc import abstractmethod
from typing import List, Type

from AnyQt.QtCore import QUrl
from AnyQt.QtGui import QDropEvent

from orangewidget.utils.filedialogs import (
    open_filename_dialog_save, open_filename_dialog,
    RecentPath, RecentPathsWidgetMixin, RecentPathsWComboMixin,
)

# imported for backcompatibility
from orangewidget.utils.filedialogs import (  # pylint: disable=unused-import
    fix_extension, format_filter, get_file_name, Compression
)

from Orange.data.io import FileFormat
from Orange.util import deprecated
from Orange.widgets.widget import OWWidget

__all__ = [
    "open_filename_dialog_save", "open_filename_dialog",
    "RecentPath", "RecentPathsWidgetMixin", "RecentPathsWComboMixin",
    "stored_recent_paths_prepend", "OWUrlDropBase"
]


@deprecated
def dialog_formats():
    """
    Return readable file types for QFileDialogs.
    """
    return ("All readable files ({});;".format(
        '*' + ' *'.join(FileFormat.readers.keys())) +
            ";;".join("{} (*{})".format(f.DESCRIPTION, ' *'.join(f.EXTENSIONS))
                      for f in sorted(set(FileFormat.readers.values()),
                                      key=list(FileFormat.readers.values()).index)))


def stored_recent_paths_prepend(
        class_: Type[RecentPathsWidgetMixin], r: RecentPath
) -> List[RecentPath]:
    """
    Load existing stored defaults *recent_paths* and move or prepend
    `r` to front.
    """
    existing = get_stored_default_recent_paths(class_)
    if r in existing:
        existing.remove(r)
    return [r] + existing


def get_stored_default_recent_paths(class_: Type[RecentPathsWidgetMixin]):
    recent_paths = []
    try:
        items = class_.settingsHandler.defaults.get("recent_paths", [])
        for item in items:
            if isinstance(item, RecentPath):
                recent_paths.append(item)
    except (AttributeError, KeyError, TypeError):
        pass
    return recent_paths


class OWUrlDropBase(OWWidget, openclass=True):
    """
    A abstract base class for a OWBaseWidget that accepts url drops.
    """
    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)
        self.setAcceptDrops(True)

    @abstractmethod
    def canDropUrl(self, url: QUrl) -> bool:
        """
        Can the `url` be dropped on this widget.

        This method must be reimplemented in a subclass.
        """
        raise NotImplementedError

    @abstractmethod
    def handleDroppedUrl(self, url: QUrl) -> None:
        """
        Handle the dropped `url`.

        This method must be reimplemented in a subclass.
        """
        raise NotImplementedError

    def dragEnterEvent(self, event):
        urls = event.mimeData().urls()
        if urls and self.canDropUrl(urls[0]):
            event.acceptProposedAction()
            return
        super().dragEnterEvent(event)

    def dragMoveEvent(self, event):
        urls = event.mimeData().urls()
        if urls and self.canDropUrl(urls[0]):
            event.acceptProposedAction()
            return
        super().dragMoveEvent(event)

    def dropEvent(self, event: QDropEvent):
        urls = event.mimeData().urls()
        if urls and self.canDropUrl(urls[0]):
            self.handleDroppedUrl(urls[0])
            event.acceptProposedAction()
            return
        super().dropEvent(event)



import sys
from typing import Optional

from AnyQt.QtCore import Signal, Qt
from AnyQt.QtWidgets import QVBoxLayout, QWidget, QComboBox, \
    QFormLayout, QLabel, QButtonGroup, QRadioButton, QLayout

from Orange.data import ContinuousVariable, Domain
from Orange.widgets.utils import disconnected
from Orange.widgets.utils.itemmodels import DomainModel


class ErrorBarsDialog(QWidget):
    changed = Signal()

    def __init__(
            self,
            parent: QWidget,
    ):
        super().__init__(parent)
        self.setWindowFlags(self.windowFlags() | Qt.Popup)
        self.hide()
        self.__model = DomainModel(
            separators=False,
            valid_types=(ContinuousVariable,),
            placeholder="(None)"
        )

        self.__upper_combo = upper_combo = QComboBox()
        upper_combo.setMinimumWidth(200)
        upper_combo.setModel(self.__model)
        upper_combo.currentIndexChanged.connect(self.changed)

        self.__lower_combo = lower_combo = QComboBox()
        lower_combo.setMinimumWidth(200)
        lower_combo.setModel(self.__model)
        lower_combo.currentIndexChanged.connect(self.changed)

        button_diff = QRadioButton("Difference from plotted value",
                                   checked=True)
        button_abs = QRadioButton("Absolute position on the plot")
        self.__radio_buttons = QButtonGroup()
        self.__radio_buttons.addButton(button_diff, 0)
        self.__radio_buttons.addButton(button_abs, 1)
        self.__radio_buttons.buttonClicked.connect(self.changed)

        form = QFormLayout()
        form.addRow(QLabel("Upper:"), upper_combo)
        form.addRow(QLabel("Lower:"), lower_combo)
        form.setVerticalSpacing(10)
        form.addRow(button_diff)
        form.addRow(button_abs)

        layout = QVBoxLayout()
        self.setLayout(layout)
        layout.addLayout(form)
        layout.setSizeConstraint(QLayout.SizeConstraint.SetFixedSize)

    def get_data(self) -> tuple[
        Optional[ContinuousVariable], Optional[ContinuousVariable], bool
    ]:
        upper_var, lower_var = None, None
        if self.__model:
            upper_var = self.__model[self.__upper_combo.currentIndex()]
            lower_var = self.__model[self.__lower_combo.currentIndex()]
        return upper_var, lower_var, bool(self.__radio_buttons.checkedId())

    def show_dlg(
            self,
            domain: Domain,
            x: int, y: int,
            attr_upper: Optional[ContinuousVariable] = None,
            attr_lower: Optional[ContinuousVariable] = None,
            is_abs: bool = True
    ):
        self._set_data(domain, attr_upper, attr_lower, is_abs)
        self.show()
        self.raise_()
        self.move(x, y)
        self.activateWindow()

    def _set_data(
            self,
            domain: Domain,
            upper_attr: Optional[ContinuousVariable],
            lower_attr: Optional[ContinuousVariable],
            is_abs: bool
    ):
        upper_combo, lower_combo = self.__upper_combo, self.__lower_combo
        with disconnected(upper_combo.currentIndexChanged, self.changed):
            with disconnected(lower_combo.currentIndexChanged, self.changed):
                self.__model.set_domain(domain)
                upper_combo.setCurrentIndex(self.__model.indexOf(upper_attr))
                lower_combo.setCurrentIndex(self.__model.indexOf(lower_attr))
        self.__radio_buttons.buttons()[int(is_abs)].setChecked(True)


if __name__ == "__main__":
    # pylint: disable=ungrouped-imports
    from AnyQt.QtWidgets import QApplication, QPushButton

    from Orange.data import Table

    app = QApplication(sys.argv)
    w = QWidget()
    w.setFixedSize(400, 200)

    dlg = ErrorBarsDialog(w)
    dlg.changed.connect(lambda: print(dlg.get_data()))

    btn = QPushButton(w)
    btn.setText("Open")

    _domain: Domain = Table("iris").domain


    def _on_click():
        dlg.show_dlg(_domain, 500, 500, _domain.attributes[2],
                     _domain.attributes[3], is_abs=False)


    btn.clicked.connect(_on_click)

    w.show()
    sys.exit(app.exec())



# pylint: disable=missing-docstring,protected-access
import unittest
from unittest.mock import patch, Mock

import pyqtgraph as pg

from AnyQt.QtCore import QPointF
from AnyQt.QtGui import QFont
from AnyQt.QtWidgets import QToolTip

from Orange.classification import NaiveBayesLearner
from Orange.data import Table, Domain
from Orange.modelling import RandomForestLearner
from Orange.regression import PLSRegressionLearner
from Orange.widgets.evaluate.owparameterfitter import OWParameterFitter
from Orange.widgets.model.owrandomforest import OWRandomForest
from Orange.widgets.tests.base import WidgetTest
from Orange.widgets.tests.utils import simulate


class DummyLearner(PLSRegressionLearner):
    @property
    def fitted_parameters(self):
        return [
            self.FittedParameter("n_components", "Foo", int, 5, None),
            self.FittedParameter("n_components", "Bar", int, 5, 10),
            self.FittedParameter("n_components", "Baz", int, None, 10),
            self.FittedParameter("n_components", "Qux", int, None, None)
        ]


class TestOWParameterFitter(WidgetTest):
    @classmethod
    def setUpClass(cls):
        super().setUpClass()
        cls._heart = Table("heart_disease")
        cls._housing = Table("housing")
        cls._naive_bayes = NaiveBayesLearner()
        cls._pls = PLSRegressionLearner()
        cls._rf = RandomForestLearner()
        cls._dummy = DummyLearner()

    def setUp(self):
        self.widget = self.create_widget(OWParameterFitter)

    def test_init(self):
        self.widget.controls.minimum.setValue(3)
        self.widget.controls.maximum.setValue(6)

        self.send_signal(self.widget.Inputs.learner, self._pls)
        self.assertEqual(self.widget.controls.parameter_index.currentText(),
                         "Components")
        self.assertEqual(self.widget.minimum, 3)
        self.assertEqual(self.widget.maximum, 6)

        self.send_signal(self.widget.Inputs.learner, None)
        self.assertEqual(self.widget.controls.parameter_index.currentText(),
                         "")

    def test_input(self):
        self.send_signal(self.widget.Inputs.data, self._housing)
        self.send_signal(self.widget.Inputs.learner, self._pls)
        self.wait_until_finished()

        self.send_signal(self.widget.Inputs.data, self._heart)
        self.wait_until_finished()
        self.assertTrue(self.widget.Error.incompatible_learner.is_shown())

        self.send_signal(self.widget.Inputs.learner, None)
        self.assertFalse(self.widget.Error.incompatible_learner.is_shown())

    def test_input_no_params(self):
        self.send_signal(self.widget.Inputs.data, self._heart)
        self.send_signal(self.widget.Inputs.learner, self._naive_bayes)
        self.wait_until_finished()
        self.assertTrue(self.widget.Warning.no_parameters.is_shown())

        self.send_signal(self.widget.Inputs.learner, None)
        self.assertFalse(self.widget.Warning.no_parameters.is_shown())

    def test_random_forest(self):
        rf_widget = self.create_widget(OWRandomForest)
        learner = self.get_output(rf_widget.Outputs.learner)

        self.send_signal(self.widget.Inputs.learner, learner)
        self.assertFalse(self.widget.Warning.no_parameters.is_shown())
        self.assertFalse(self.widget.Error.unknown_err.is_shown())
        self.assertFalse(self.widget.Error.not_enough_data.is_shown())
        self.assertFalse(self.widget.Error.incompatible_learner.is_shown())

        self.send_signal(self.widget.Inputs.data, self._heart)
        self.assertFalse(self.widget.Warning.no_parameters.is_shown())
        self.assertFalse(self.widget.Error.unknown_err.is_shown())
        self.assertFalse(self.widget.Error.not_enough_data.is_shown())
        self.assertFalse(self.widget.Error.incompatible_learner.is_shown())

        self.send_signal(self.widget.Inputs.data, self._housing)
        self.assertFalse(self.widget.Warning.no_parameters.is_shown())
        self.assertFalse(self.widget.Error.unknown_err.is_shown())
        self.assertFalse(self.widget.Error.not_enough_data.is_shown())
        self.assertFalse(self.widget.Error.incompatible_learner.is_shown())

    def test_classless_data(self):
        data = self._housing
        classless_data = data.transform(Domain(data.domain.attributes))

        self.send_signal(self.widget.Inputs.learner, self._pls)
        self.send_signal(self.widget.Inputs.data, classless_data)
        self.wait_until_finished()
        self.assertTrue(self.widget.Error.missing_target.is_shown())

        self.send_signal(self.widget.Inputs.data, data)
        self.wait_until_finished()
        self.assertFalse(self.widget.Error.missing_target.is_shown())

    def test_multiclass_data(self):
        data = self._housing
        multiclass_data = data.transform(Domain(data.domain.attributes[2:],
                                                data.domain.attributes[:2]))

        self.send_signal(self.widget.Inputs.learner, self._pls)
        self.send_signal(self.widget.Inputs.data, multiclass_data)
        self.wait_until_finished()
        self.assertTrue(self.widget.Error.multiple_targets_data.is_shown())

        self.send_signal(self.widget.Inputs.data, data)
        self.wait_until_finished()
        self.assertFalse(self.widget.Error.multiple_targets_data.is_shown())

    def test_plot(self):
        self.send_signal(self.widget.Inputs.data, self._housing)
        self.send_signal(self.widget.Inputs.learner, self._pls)
        self.wait_until_finished()

        x = self.widget.graph._FitterPlot__bar_item_tr.opts["x"]
        self.assertEqual(list(x), [-0.2, 0.8])
        x = self.widget.graph._FitterPlot__bar_item_cv.opts["x"]
        self.assertEqual(list(x), [0.2, 1.2])

    @patch.object(QToolTip, "showText")
    def test_tooltip(self, show_text):
        graph = self.widget.graph

        self.assertFalse(self.widget.graph.help_event(Mock()))
        self.assertIsNone(show_text.call_args)

        self.send_signal(self.widget.Inputs.data, self._housing)
        self.send_signal(self.widget.Inputs.learner, self._pls)
        self.wait_until_finished()

        for item in graph.items():
            if isinstance(item, pg.BarGraphItem):
                item.mapFromScene = Mock(return_value=QPointF(0.2, 0.2))

        self.assertTrue(self.widget.graph.help_event(Mock()))
        self.assertIn("Train:", show_text.call_args[0][1])
        self.assertIn("CV:", show_text.call_args[0][1])

        for item in graph.items():
            if isinstance(item, pg.BarGraphItem):
                item.mapFromScene = Mock(return_value=QPointF(0.5, 0.5))
        self.assertFalse(self.widget.graph.help_event(Mock()))

    def test_manual_steps(self):
        self.send_signal(self.widget.Inputs.data, self._housing)
        self.send_signal(self.widget.Inputs.learner, self._pls)
        self.wait_until_finished()

        self.widget.controls.manual_steps.setText("1, 2, 3")
        self.widget.controls.type.buttons[1].click()
        self.wait_until_finished()

        x = self.widget.graph._FitterPlot__bar_item_tr.opts["x"]
        self.assertEqual(list(x), [-0.2, 0.8, 1.8])
        x = self.widget.graph._FitterPlot__bar_item_cv.opts["x"]
        self.assertEqual(list(x), [0.2, 1.2, 2.2])

    def test_manual_steps_limits(self):
        w = self.widget

        def check(cases):
            for setting, steps in cases:
                w.controls.manual_steps.setText(setting)
                w.controls.manual_steps.returnPressed.emit()
                self.assertEqual(w.steps, steps, f"setting: {setting}")
                self.assertIs(w.Error.manual_steps_error.is_shown(), not steps,
                              f"setting: {setting}")

        self.send_signal(w.Inputs.data, self._housing)
        self.send_signal(w.Inputs.learner, self._dummy)
        self.wait_until_finished()

        # 5 to None
        simulate.combobox_activate_index(w.controls.parameter_index, 0)
        self.wait_until_finished()
        check([("6, 9, 7", (6, 7, 9)),
               ("6, 9, 7, 3", ()),
               ("6, 9, 7", (6, 7, 9)),
               ("6, 9, 7, 3", ())])

        # None to 10
        simulate.combobox_activate_index(w.controls.parameter_index, 2)
        self.wait_until_finished()
        self.assertFalse(w.Error.manual_steps_error.is_shown())

        check([("12, 1, 3, -5", ()),
               ("1, 3, -5", (-5, 1, 3)),
               ("12, 1, 3, -5", ())])

        # No limits
        simulate.combobox_activate_index(w.controls.parameter_index, 3)
        self.wait_until_finished()

        self.assertEqual(w.steps, (-5, 1, 3, 12))
        self.assertFalse(w.Error.manual_steps_error.is_shown())

        # 5 to 10
        simulate.combobox_activate_index(w.controls.parameter_index, 1)
        self.wait_until_finished()

        self.assertEqual(w.steps, ())
        self.assertTrue(w.Error.manual_steps_error.is_shown())

        check([("12, 8, 7, 5", ()),
               ("8, 7, -5", ()),
               ("8, 7, 5", (5, 7, 8))])

    def test_steps_preview(self):
        self.send_signal(self.widget.Inputs.data, self._housing)
        self.send_signal(self.widget.Inputs.learner, self._pls)
        self.wait_until_finished()
        self.assertEqual(self.widget.range_preview.steps(), (1, 2))

        self.widget.controls.type.buttons[1].click()
        self.wait_until_finished()
        self.assertIsNone(self.widget.range_preview.steps())

    def test_on_parameter_changed(self):
        self.send_signal(self.widget.Inputs.data, self._housing)
        self.send_signal(self.widget.Inputs.learner, self._dummy)
        self.wait_until_finished()

        self.widget.commit.deferred = Mock()

        for i in range(1, 4):
            self.widget.commit.deferred.reset_mock()
            simulate.combobox_activate_index(
                self.widget.controls.parameter_index, i)
            self.wait_until_finished()
            self.widget.commit.deferred.assert_called_once()

    def test_not_enough_data(self):
        self.send_signal(self.widget.Inputs.data, self._housing[:5])
        self.send_signal(self.widget.Inputs.learner, self._pls)
        self.wait_until_finished()
        self.assertTrue(self.widget.Error.not_enough_data.is_shown())
        self.send_signal(self.widget.Inputs.data, None)
        self.assertFalse(self.widget.Error.not_enough_data.is_shown())

    def test_unknown_err(self):
        self.send_signal(self.widget.Inputs.data, Table("iris")[:50])
        self.send_signal(self.widget.Inputs.learner, self._rf)
        self.wait_until_finished()
        self.assertTrue(self.widget.Error.unknown_err.is_shown())
        self.send_signal(self.widget.Inputs.data, None)
        self.assertFalse(self.widget.Error.unknown_err.is_shown())

    def test_fitted_parameters(self):
        self.assertEqual(self.widget.fitted_parameters, [])

        self.send_signal(self.widget.Inputs.data, self._housing)
        self.assertEqual(self.widget.fitted_parameters, [])

        self.send_signal(self.widget.Inputs.learner, self._pls)
        self.assertEqual(len(self.widget.fitted_parameters), 1)
        self.wait_until_finished()

    def test_initial_parameters(self):
        self.assertEqual(self.widget.initial_parameters, {})

        self.send_signal(self.widget.Inputs.data, self._housing)
        self.assertEqual(self.widget.initial_parameters, {})

        self.send_signal(self.widget.Inputs.learner, self._pls)
        self.assertEqual(len(self.widget.initial_parameters), 3)
        self.wait_until_finished()

        self.send_signal(self.widget.Inputs.learner, self._rf)
        self.assertEqual(len(self.widget.initial_parameters), 13)
        self.wait_until_finished()

        self.send_signal(self.widget.Inputs.data, None)
        self.assertEqual(len(self.widget.initial_parameters), 14)

        self.send_signal(self.widget.Inputs.learner, None)
        self.assertEqual(self.widget.initial_parameters, {})

    def test_bounds(self):
        self.widget.controls.minimum.setValue(-3)
        self.widget.controls.maximum.setValue(6)
        self.send_signal(self.widget.Inputs.learner, self._pls)
        self.send_signal(self.widget.Inputs.data, self._housing)
        self.send_signal(self.widget.Inputs.learner, None)
        self.widget.controls.minimum.setValue(-3)
        self.widget.controls.maximum.setValue(6)
        self.send_signal(self.widget.Inputs.learner, self._pls)
        self.wait_until_finished()
        self.assertFalse(self.widget.Error.unknown_err.is_shown())

    def test_saved_workflow(self):
        self.send_signal(self.widget.Inputs.data, self._housing)
        self.send_signal(self.widget.Inputs.learner, self._dummy)
        self.wait_until_finished()
        simulate.combobox_activate_index(
            self.widget.controls.parameter_index, 2)
        self.widget.controls.minimum.setValue(3)
        self.widget.controls.maximum.setValue(6)
        self.wait_until_finished()

        settings = self.widget.settingsHandler.pack_data(self.widget)
        widget = self.create_widget(OWParameterFitter,
                                    stored_settings=settings)
        self.send_signal(widget.Inputs.data, self._housing, widget=widget)
        self.send_signal(widget.Inputs.learner, self._dummy, widget=widget)
        self.wait_until_finished(widget=widget)
        self.assertEqual(widget.controls.parameter_index.currentText(), "Baz")
        self.assertEqual(widget.minimum, 3)
        self.assertEqual(widget.maximum, 6)

    def test_retain_settings(self):
        self.send_signal(self.widget.Inputs.learner, self._dummy)

        controls = self.widget.controls

        def _test():
            self.assertEqual(controls.parameter_index.currentText(), "Bar")
            self.assertEqual(controls.minimum.value(), 6)
            self.assertEqual(controls.maximum.value(), 8)
            self.assertEqual(self.widget.parameter_index, 1)
            self.assertEqual(self.widget.minimum, 6)
            self.assertEqual(self.widget.maximum, 8)

        simulate.combobox_activate_index(controls.parameter_index, 1)
        controls.minimum.setValue(6)
        controls.maximum.setValue(8)

        self.send_signal(self.widget.Inputs.data, self._housing)
        _test()

        self.send_signal(self.widget.Inputs.learner,
                         DummyLearner(n_components=6))
        _test()

        self.send_signal(self.widget.Inputs.data, None)
        self.send_signal(self.widget.Inputs.data, self._housing)
        _test()

        self.send_signal(self.widget.Inputs.learner, self._rf)
        self.assertEqual(controls.parameter_index.currentText(),
                         "Number of trees")
        self.assertEqual(controls.minimum.value(), 1)
        self.assertEqual(controls.maximum.value(), 10)
        self.assertEqual(self.widget.parameter_index, 0)
        self.assertEqual(self.widget.minimum, 1)
        self.assertEqual(self.widget.maximum, 10)

    def test_visual_settings(self):
        graph = self.widget.graph

        def test_settings():
            font = QFont("Helvetica", italic=True, pointSize=20)
            for item in graph.parameter_setter.axis_items:
                self.assertFontEqual(item.label.font(), font)
            font.setPointSize(15)
            for item in graph.parameter_setter.axis_items:
                self.assertFontEqual(item.style["tickFont"], font)
            font.setPointSize(17)
            for legend_item in graph.parameter_setter.legend_items:
                self.assertFontEqual(legend_item[1].item.font(), font)
            self.assertFalse(graph.getAxis("left").grid)

        key, value = ("Fonts", "Font family", "Font family"), "Helvetica"
        self.widget.set_visual_settings(key, value)

        key, value = ("Fonts", "Axis title", "Font size"), 20
        self.widget.set_visual_settings(key, value)
        key, value = ("Fonts", "Axis title", "Italic"), True
        self.widget.set_visual_settings(key, value)

        key, value = ("Fonts", "Axis ticks", "Font size"), 15
        self.widget.set_visual_settings(key, value)
        key, value = ("Fonts", "Axis ticks", "Italic"), True
        self.widget.set_visual_settings(key, value)

        key, value = ("Fonts", "Legend", "Font size"), 17
        self.widget.set_visual_settings(key, value)
        key, value = ("Fonts", "Legend", "Italic"), True
        self.widget.set_visual_settings(key, value)

        key, value = ("Figure", "Gridlines", "Show"), False
        self.widget.set_visual_settings(key, value)
        key, value = ("Figure", "Gridlines", "Opacity"), 20
        self.widget.set_visual_settings(key, value)

        test_settings()

        self.send_signal(self.widget.Inputs.learner, self._pls)
        self.send_signal(self.widget.Inputs.data, self._heart[:10])
        test_settings()

        self.send_signal(self.widget.Inputs.data, None)
        self.send_signal(self.widget.Inputs.learner, None)

        self.send_signal(self.widget.Inputs.learner, self._pls)
        self.send_signal(self.widget.Inputs.data, self._heart[:10])
        test_settings()

    def assertFontEqual(self, font1: QFont, font2: QFont):
        self.assertEqual(font1.family(), font2.family())
        self.assertEqual(font1.pointSize(), font2.pointSize())
        self.assertEqual(font1.italic(), font2.italic())

    def test_send_report(self):
        self.widget.send_report()

        self.send_signal(self.widget.Inputs.data, self._housing)
        self.send_signal(self.widget.Inputs.learner, self._pls)
        self.wait_until_finished()
        self.widget.send_report()

        self.send_signal(self.widget.Inputs.data, self._heart)
        self.send_signal(self.widget.Inputs.learner, self._naive_bayes)
        self.wait_until_finished()
        self.widget.send_report()

    def test_steps_from_range_error(self):
        w: OWParameterFitter = self.widget
        self.send_signal(w.Inputs.data, self._heart)
        self.send_signal(w.Inputs.learner, self._dummy)
        w.type = w.FROM_RANGE

        w.minimum = 10
        w.maximum = 5
        self.assertEqual(w.steps, ())
        self.assertTrue(w.Error.min_max_error.is_shown())

        w.maximum = 15
        self.assertNotEqual(w.steps, ())
        self.assertFalse(w.Error.min_max_error.is_shown())

        w.minimum = 10
        w.maximum = 5
        w.steps  # pylint: disable=pointless-statement
        self.assertTrue(w.Error.min_max_error.is_shown())

        self.send_signal(w.Inputs.learner, None)
        self.assertFalse(w.Error.min_max_error.is_shown())

    def test_steps_from_ranges_steps(self):
        w: OWParameterFitter = self.widget
        self.send_signal(w.Inputs.data, self._heart)
        self.send_signal(w.Inputs.learner, self._dummy)
        w.type = w.FROM_RANGE

        for mini, maxi, exp in [
            (1, 2, (1, 2)),
            (1, 5, (1, 2, 3, 4, 5)),
            (1, 10, (1, 2, 3, 4, 5, 6, 7, 8, 9, 10)),
            (2, 14, (2, 3, 4, 5, 6, 7, 8, 9, 10, 11, 12, 13, 14)),
            (2, 20, (2, 10, 20)),
            (2, 22, (2, 10, 20, 22)),
            (2, 10, (2, 3, 4, 5, 6, 7, 8, 9, 10)),
            (2, 5, (2, 3, 4, 5)),
            (2, 4, (2, 3, 4)),
            (1, 1, (1,)),
            (1, 50, (1, 10, 20, 30, 40, 50)),
            (3, 49, (3, 10, 20, 30, 40, 49)),
            (9, 31, (9, 10, 20, 30, 31)),
            (90, 398, (90, 100, 200, 300, 398)),
            (90, 1010,
             (90, 100, 200, 300, 400, 500, 600, 700, 800, 900, 1000, 1010)),
            (810, 1234, (810, 900, 1000, 1100, 1200, 1234)),
            (4980, 18030,
             (4980, 5000, 6000, 7000, 8000, 9000, 10000, 11000, 12000,
              13000, 14000, 15000, 16000, 17000, 18000, 18030))]:
            w.minimum = mini
            w.maximum = maxi
            self.assertEqual(w.steps, exp, f"min={mini}, max={maxi}")

    def test_steps_from_manual_error(self):
        w: OWParameterFitter = self.widget
        self.send_signal(w.Inputs.data, self._housing)
        self.send_signal(w.Inputs.learner, self._dummy)
        self.wait_until_finished()
        simulate.combobox_activate_index(w.controls.parameter_index, 3)
        w.type = w.MANUAL

        w.manual_steps = "1, 2, 3, asdf, 4, 5"
        self.assertEqual(w.steps, ())
        self.assertTrue(w.Error.manual_steps_error.is_shown())

        w.manual_steps = "1, 2, 3, 4, 5"
        self.assertNotEqual(w.steps, ())
        self.assertFalse(w.Error.manual_steps_error.is_shown())

        w.manual_steps = "1, 2, 3, asdf, 4, 5"
        w.steps  # pylint: disable=pointless-statement
        self.assertTrue(w.Error.manual_steps_error.is_shown())

        self.send_signal(w.Inputs.learner, None)
        self.assertFalse(w.Error.manual_steps_error.is_shown())

    def test_steps_from_manual_no_dots(self):
        w: OWParameterFitter = self.widget
        self.send_signal(w.Inputs.data, self._housing)
        self.send_signal(w.Inputs.learner, self._dummy)
        self.wait_until_finished()
        simulate.combobox_activate_index(w.controls.parameter_index, 3)
        w.type = w.MANUAL

        w.manual_steps = "1, 2, 3, 4, 5"
        self.assertEqual(w.steps, (1, 2, 3, 4, 5))

        w.manual_steps = "1, 2, 3, 4, 5, 6"
        self.assertEqual(w.steps, (1, 2, 3, 4, 5, 6))

        w.manual_steps = "1, 2, 10,   3, 4, 123, 5, 6"
        self.assertEqual(w.steps, (1, 2, 3, 4, 5, 6, 10, 123))

    def test_steps_from_manual_dots(self):
        def check(cases):
            for settings, steps in cases:
                w.manual_steps = settings
                self.assertEqual(w.steps, steps, f"setting: {settings}")
                self.assertIs(w.Error.manual_steps_error.is_shown(), not steps,
                              f"setting: {settings}")

        w: OWParameterFitter = self.widget
        self.send_signal(w.Inputs.data, self._housing)
        self.send_signal(w.Inputs.learner, self._dummy)
        self.wait_until_finished()
        w.type = w.MANUAL

        # No limits
        simulate.combobox_activate_index(w.controls.parameter_index, 3)
        self.wait_until_finished()
        check([("1, 2, ..., 5", (1, 2, 3, 4, 5)),
               ("1, 2, 3, ..., 5, 6, 7", (1, 2, 3, 4, 5, 6, 7)),
               ("3, ..., 5, 6", (3, 4, 5, 6)),
               ("..., 5, 6", ()),
               ("5, 6, ...", ()),
               ("1, 2, 3, 4, 5, ...", ()),
               ("1, ..., 5", ()),
               ("1, 2, ..., 5, 6, ..., 8", ())])

        # 5 to 10
        simulate.combobox_activate_index(w.controls.parameter_index, 1)
        self.wait_until_finished()
        check([("4, 5, ..., 8", ()),
               ("5, 6, ..., 12", ()),
               ("5, 6, ..., 9", (5, 6, 7, 8, 9)),
               ("6, 7, ..., 9", (6, 7, 8, 9)),
               ("6, 7, ..., 8, 9", (6, 7, 8, 9)),
               ("..., 8, 9", (5, 6, 7, 8, 9)),
               ("6, 7, ...", (6, 7, 8, 9, 10)),
               ("6, 7, 8, 9, ...", (6, 7, 8, 9, 10)),
               ("..., 8, 9", (5, 6, 7, 8, 9)),
               ])

        # 5 to None
        simulate.combobox_activate_index(w.controls.parameter_index, 0)
        self.wait_until_finished()
        check([("4, 5, ..., 8", ()),
               ("5, 6, ..., 12", (5, 6, 7, 8, 9, 10, 11, 12)),
               ("6, 7, ..., 9", (6, 7, 8, 9)),
               ("6, 7, ..., 8, 9", (6, 7, 8, 9)),
               ("..., 8, 9", (5, 6, 7, 8, 9)),
               ("6, 7, ...", ())
               ])

        # None to 10
        simulate.combobox_activate_index(w.controls.parameter_index, 2)
        self.wait_until_finished()
        check([("4, 5, ..., 8", (4, 5, 6, 7, 8)),
               ("5, 6, ..., 12", ()),
               ("5, 6, ..., 9", (5, 6, 7, 8, 9)),
               ("6, 7, ..., 9", (6, 7, 8, 9)),
               ("..., 8, 9", ()),
               ("6, 7, ...", (6, 7, 8, 9, 10)),
               ("6, 7, 8, 9, ...", (6, 7, 8, 9, 10)),
               ("..., 8, 9", ()),
               ])

    def test_steps_from_manual_dots_corrections(self):
        w: OWParameterFitter = self.widget
        self.send_signal(w.Inputs.data, self._housing)
        self.send_signal(w.Inputs.learner, self._dummy)
        self.wait_until_finished()
        w.type = w.MANUAL

        # 5 to 10
        simulate.combobox_activate_index(w.controls.parameter_index, 1)
        self.wait_until_finished()

        for settings, steps in [
            ("5, 6..., 8", (5, 6, 7, 8)),
            ("5,6...,8", (5, 6, 7, 8)),
            ("5,6...8", (5, 6, 7, 8)),
            ("5,   6  ...  8", (5, 6, 7, 8)),
            ("5,   6  ...  8", (5, 6, 7, 8)),
            ("5,   6  ...  ", (5, 6, 7, 8, 9, 10)),
            ("..., 7, 8", (5, 6, 7, 8)),
            ("..., 7, 8, ...", ()),
            ("5, 6, ..., 7, 8, ...", ()),
            ("5, 6, 8, ...", ()),
            ("5, 6, 8, ...", ()),
            ("5, 6, ..., 8, 10", ()),
            ("5, 7, ..., 8, 10", ()),
            ("8, 7, 6, ...", ()),
            ("5, 6, 7, ..., 7, 8", ()),
        ]:
            w.manual_steps = settings
            self.assertEqual(w.steps, steps, f"setting: {settings}")
            self.assertIs(w.Error.manual_steps_error.is_shown(), not steps,
                          f"setting: {settings}")

    def test_manual_tooltip(self):
        w: OWParameterFitter = self.widget
        self.send_signal(w.Inputs.data, self._housing)
        self.send_signal(w.Inputs.learner, self._dummy)
        self.wait_until_finished()

        simulate.combobox_activate_index(w.controls.parameter_index, 0)
        self.assertIn("greater or equal to 5", w.edit.toolTip())

        simulate.combobox_activate_index(w.controls.parameter_index, 1)
        self.assertIn("between 5 and 10", w.edit.toolTip())

        simulate.combobox_activate_index(w.controls.parameter_index, 2)
        self.assertIn("smaller or equal to 10", w.edit.toolTip())

        simulate.combobox_activate_index(w.controls.parameter_index, 3)
        self.assertEqual("", w.edit.toolTip())


if __name__ == "__main__":
    unittest.main()



import unittest

from orangewidget.tests.base import WidgetTest

from Orange.data import Table
from Orange.preprocess import Impute

from Orange.classification.scoringsheet import ScoringSheetLearner
from Orange.widgets.model.owscoringsheet import OWScoringSheet


class TestOWScoringSheet(WidgetTest):
    @classmethod
    def setUpClass(cls):
        super().setUpClass()
        cls.heart = Table("heart_disease")
        cls.housing = Table("housing")
        cls.scoring_sheet_learner = ScoringSheetLearner(20, 5, 5, None)
        cls.scoring_sheet_model = cls.scoring_sheet_learner(cls.heart)

    def setUp(self):
        self.widget = self.create_widget(OWScoringSheet)

    def test_no_data_input(self):
        self.assertIsNotNone(self.get_output(self.widget.Outputs.learner))
        self.assertIsNone(self.get_output(self.widget.Outputs.model))

    def test_numerical_target_attribute(self):
        self.send_signal(self.widget.Inputs.data, self.housing)
        self.wait_until_finished()
        self.assertTrue(self.widget.Error.fitting_failed.is_shown())

    def test_settings_in_learner(self):
        self.widget.num_attr_after_selection = 20
        self.widget.num_decision_params = 7
        self.widget.max_points_per_param = 8
        self.widget.custom_features_checkbox = True
        self.widget.num_input_features = 4

        self.widget.apply()

        self.send_signal(self.widget.Inputs.data, self.heart)
        learner = self.get_output(self.widget.Outputs.learner)

        self.assertEqual(learner.num_decision_params, 7)
        self.assertEqual(learner.max_points_per_param, 8)
        self.assertEqual(learner.num_input_features, 4)

    def test_settings_in_model(self):
        self.widget.num_attr_after_selection = 20
        self.widget.num_decision_params = 7
        self.widget.max_points_per_param = 8
        self.widget.custom_features_checkbox = True
        self.widget.num_input_features = 4

        self.widget.apply()

        self.send_signal(self.widget.Inputs.data, self.heart)
        self.wait_until_finished()
        model = self.get_output(self.widget.Outputs.model)

        coefficients = model.model.coefficients
        non_zero_coefficients = [coef for coef in coefficients if coef != 0]

        self.assertEqual(len(coefficients), self.widget.num_attr_after_selection)

        # most often equal, but in some cases the optimizer finds fewer parameters
        self.assertLessEqual(len(non_zero_coefficients), self.widget.num_decision_params)

        self.assertLessEqual(
            max(non_zero_coefficients, key=lambda x: abs(x)),
            self.widget.max_points_per_param,
        )

    def test_custom_number_input_features_information(self):
        self.widget.custom_features_checkbox = True
        self.widget.custom_input_features()
        self.assertTrue(self.widget.Information.custom_num_of_input_features.is_shown())

        self.widget.custom_features_checkbox = False
        self.widget.custom_input_features()
        self.assertFalse(
            self.widget.Information.custom_num_of_input_features.is_shown()
        )

    def test_custom_preprocessors_information(self):
        preprocessor = Impute()
        self.send_signal(self.widget.Inputs.preprocessor, preprocessor)
        self.assertTrue(self.widget.Information.ignored_preprocessors.is_shown())

        self.send_signal(self.widget.Inputs.preprocessor, None)
        self.assertFalse(self.widget.Information.ignored_preprocessors.is_shown())

    def test_custom_preprocessors_spin_disabled(self):
        preprocessor = Impute()
        self.send_signal(self.widget.Inputs.preprocessor, preprocessor)
        self.assertFalse(self.widget.num_attr_after_selection_spin.isEnabled())

    def test_default_preprocessors_are_used(self):
        learner = self.get_output(self.widget.Outputs.learner)

        self.assertIsNotNone(learner.preprocessors)
        self.assertEqual(len(learner.preprocessors), 5)

    def test_custom_preprocessors_are_used(self):
        preprocessor = Impute()
        self.send_signal(self.widget.Inputs.preprocessor, preprocessor)
        learner = self.get_output(self.widget.Outputs.learner)

        self.assertIsNotNone(learner.preprocessors)
        self.assertEqual(len(learner.preprocessors), 1)
        self.assertEqual(learner.preprocessors[0], preprocessor)


if __name__ == "__main__":
    unittest.main()



import numpy as np
import sys
# import warnings
# warnings.filterwarnings("ignore")
from Orange.classification.utils.fasterrisk.utils import get_support_indices, get_nonsupport_indices, compute_logisticLoss_from_ExpyXB
from Orange.classification.utils.fasterrisk.base_model import logRegModel

class sparseDiversePoolLogRegModel(logRegModel):
    def __init__(self, X, y, lambda2=1e-8, intercept=True, original_lb=-5, original_ub=5):
        super().__init__(X=X, y=y, lambda2=lambda2, intercept=intercept, original_lb=original_lb, original_ub=original_ub)
   
    def getAvailableIndices_for_expansion_but_avoid_l(self, nonsupport, support, l):
        """Get the indices of features that can be added to the support of the current sparse solution

        Parameters
        ----------
        betas : ndarray
            (1D array with `float` type) The current sparse solution

        Returns
        -------
        available_indices : ndarray
            (1D array with `int` type) The indices of features that can be added to the support of the current sparse solution
        """
        return nonsupport

    def get_sparseDiversePool(self, gap_tolerance=0.05, select_top_m=10, maxAttempts=50):
        """For the current sparse solution, get from the sparse diverse pool [select_top_m] solutions, which perform equally well as the current sparse solution. This sparse diverse pool is also called the Rashomon set. We discover new solutions by swapping 1 feature in the support of the current sparse solution.

        Parameters
        ----------
        gap_tolerance : float, optional
            New solution is accepted after swapping features if the new loss is within the [gap_tolerance] of the old loss, by default 0.05
        select_top_m : int, optional
            We select the top [select_top_m] solutions from support_size*maxAttempts number of new solutions, by default 10
        maxAttempts : int, optional
            We try to swap each feature in the support with [maxAttempts] of new features, by default 50

        Returns
        -------
        intercept_array : ndarray
            (1D array with `float` type) Return the intercept array with shape = (select_top_m, )
        coefficients_array : ndarray
            (2D array with `float` type) Return the coefficients array with shape = (select_top_m, p)
        """
        # select top m solutions with the lowest logistic losses
        # Note Bene: loss comparison here does not include logistic loss
        nonzero_indices = get_support_indices(self.betas)
        zero_indices = get_nonsupport_indices(self.betas)

        num_support = len(nonzero_indices)
        num_nonsupport = len(zero_indices)

        maxAttempts = min(maxAttempts, num_nonsupport)
        max_num_new_js = maxAttempts

        total_solutions = 1 + num_support * maxAttempts
        sparseDiversePool_betas = np.zeros((total_solutions, self.p))
        sparseDiversePool_betas[:, nonzero_indices] = self.betas[nonzero_indices]

        sparseDiversePool_beta0 = self.beta0 * np.ones((total_solutions, ))
        sparseDiversePool_ExpyXB = np.zeros((total_solutions, self.n))
        sparseDiversePool_loss = 1e12 * np.ones((total_solutions, ))

        sparseDiversePool_ExpyXB[-1] = self.ExpyXB
        sparseDiversePool_loss[-1] = compute_logisticLoss_from_ExpyXB(self.ExpyXB) + self.lambda2 * self.betas[nonzero_indices].dot(self.betas[nonzero_indices])

        betas_squareSum = self.betas[nonzero_indices].dot(self.betas[nonzero_indices])

        totalNum_in_diverseSet = 1
        for num_old_j, old_j in enumerate(nonzero_indices):
            # pick $maxAttempt$ number of features that can replace old_j
            sparseDiversePool_start = num_old_j * maxAttempts
            sparseDiversePool_end = (1 + num_old_j) * maxAttempts

            sparseDiversePool_ExpyXB[sparseDiversePool_start:sparseDiversePool_end] = self.ExpyXB * np.exp(-self.yXT[old_j] * self.betas[old_j])

            sparseDiversePool_betas[sparseDiversePool_start:sparseDiversePool_end, old_j] = 0
            
            betas_no_old_j_squareSum = betas_squareSum - self.betas[old_j]**2

            availableIndices = self.getAvailableIndices_for_expansion_but_avoid_l(zero_indices, nonzero_indices, old_j) 

            grad_on_availableIndices = -self.yXT[availableIndices].dot(np.reciprocal(1+sparseDiversePool_ExpyXB[sparseDiversePool_start]))
            abs_grad_on_availableIndices = np.abs(grad_on_availableIndices)

            # new_js = np.argpartition(abs_full_grad, -max_num_new_js)[-max_num_new_js:]
            new_js = availableIndices[np.argsort(-abs_grad_on_availableIndices)[:max_num_new_js]]

            for num_new_j, new_j in enumerate(new_js):
                sparseDiversePool_index = sparseDiversePool_start + num_new_j
                for _ in range(10):
                    self.optimize_1step_at_coord(sparseDiversePool_ExpyXB[sparseDiversePool_index], sparseDiversePool_betas[sparseDiversePool_index], self.yXT[new_j, :], new_j)
                
                loss_sparseDiversePool_index = compute_logisticLoss_from_ExpyXB(sparseDiversePool_ExpyXB[sparseDiversePool_index]) + self.lambda2 * (betas_no_old_j_squareSum + sparseDiversePool_betas[sparseDiversePool_index, new_j] ** 2)

                if (loss_sparseDiversePool_index - sparseDiversePool_loss[-1]) / sparseDiversePool_loss[-1] < gap_tolerance:
                    totalNum_in_diverseSet += 1

                    sparseDiversePool_ExpyXB[sparseDiversePool_index], sparseDiversePool_beta0[sparseDiversePool_index], sparseDiversePool_betas[sparseDiversePool_index] = self.finetune_on_current_support(sparseDiversePool_ExpyXB[sparseDiversePool_index], sparseDiversePool_beta0[sparseDiversePool_index], sparseDiversePool_betas[sparseDiversePool_index])

                    sparseDiversePool_loss[sparseDiversePool_index] = compute_logisticLoss_from_ExpyXB(sparseDiversePool_ExpyXB[sparseDiversePool_index]) + self.lambda2 * (betas_no_old_j_squareSum + sparseDiversePool_betas[sparseDiversePool_index, new_j] ** 2)

        selected_sparseDiversePool_indices = np.argsort(sparseDiversePool_loss)[:totalNum_in_diverseSet][:select_top_m]

        top_m_original_betas = np.zeros((len(selected_sparseDiversePool_indices), self.p))
        top_m_original_betas[:, self.scaled_feature_indices] = sparseDiversePool_betas[selected_sparseDiversePool_indices][:, self.scaled_feature_indices] / self.X_norm[self.scaled_feature_indices]
        top_m_original_beta0 = sparseDiversePool_beta0[selected_sparseDiversePool_indices] - top_m_original_betas.dot(self.X_mean)

        return top_m_original_beta0, top_m_original_betas

        original_sparseDiversePool_solution[1:] = sparseDiversePool_betas[selected_sparseDiversePool_indices].T
        original_sparseDiversePool_solution[1+self.scaled_feature_indices] /= self.X_norm[self.scaled_feature_indices].reshape(-1, 1)

        original_sparseDiversePool_solution[0] = sparseDiversePool_beta0[selected_sparseDiversePool_indices]
        original_sparseDiversePool_solution[0] -= self.X_mean.T @ original_sparseDiversePool_solution[1:]
        
        return original_sparseDiversePool_solution # (1+p, m) m is the number of solutions in the pool

class groupSparseDiversePoolLogRegModel(sparseDiversePoolLogRegModel):
    def __init__(self, X, y, lambda2=1e-8, intercept=True, original_lb=-5, original_ub=5, group_sparsity=10, featureIndex_to_groupIndex=None, groupIndex_to_featureIndices=None):
        super().__init__(X=X, y=y, lambda2=lambda2, intercept=intercept, original_lb=original_lb, original_ub=original_ub)

        self.group_sparsity = group_sparsity
        self.featureIndex_to_groupIndex = featureIndex_to_groupIndex
        self.groupIndex_to_featureIndices = groupIndex_to_featureIndices
    
    def getAvailableIndices_for_expansion_but_avoid_l(self, nonsupport, support, l):
        """Get the indices of features that can be added to the support of the current sparse solution

        Parameters
        ----------
        nonsupport : ndarray
            (1D array with `int` type) The indices of features that are not in the support of the current sparse solution
        support : ndarray
            (1D array with `int` type) The indices of features that are in the support of the current sparse solution
        l : int
            The index of the feature that is to be removed from the support of the current sparse solution and this index l belongs to support

        Returns
        -------
        available_indices : ndarray
            (1D array with `int` type) The indices of features that can be added to the support of the current sparse solution when we delete index l
        """
        existing_groupIndices, freq_existing_groupIndices = np.unique(self.featureIndex_to_groupIndex[support], return_counts=True)
        freq_groupIndex_of_l = freq_existing_groupIndices[existing_groupIndices == self.featureIndex_to_groupIndex[l]]
        if len(existing_groupIndices) < self.group_sparsity:
            # we have not reached the group size yet 
            available_indices = nonsupport
        elif freq_groupIndex_of_l == 1:
            # or if we remove index l, we still do not reach the group size
            available_indices = nonsupport
        else:
            # we reach the group size even if we remove index l
            available_indices = set()
            for groupIndex in existing_groupIndices:
                available_indices.update(self.groupIndex_to_featureIndices[groupIndex])
            available_indices = available_indices - set(support)
            available_indices = np.array(list(available_indices), dtype=int)

        return available_indices



import numpy as np
from itertools import product
import requests

def get_groupIndex_to_featureIndices(featureIndex_to_groupIndex):
    groupIndex_to_featureIndices = {}
    for featureIndex, groupIndex in enumerate(featureIndex_to_groupIndex):
        if groupIndex not in groupIndex_to_featureIndices:
            groupIndex_to_featureIndices[groupIndex] = set()
        groupIndex_to_featureIndices[groupIndex].add(featureIndex)
    return groupIndex_to_featureIndices

def get_support_indices(betas):
    return np.where(np.abs(betas) > 1e-9)[0]

def get_nonsupport_indices(betas):
    return np.where(np.abs(betas) <= 1e-9)[0]

def normalize_X(X):
    X_mean = np.mean(X, axis=0)
    X_norm = np.linalg.norm(X-X_mean, axis=0)
    scaled_feature_indices = np.where(X_norm >= 1e-9)[0]
    X_normalized = X-X_mean
    X_normalized[:, scaled_feature_indices] = X_normalized[:, scaled_feature_indices]/X_norm[[scaled_feature_indices]]
    return X_normalized, X_mean, X_norm, scaled_feature_indices

def compute_logisticLoss_from_yXB(yXB):
    # shape of yXB is (n, )
    return np.sum(np.log(1.+np.exp(-yXB)))

def compute_logisticLoss_from_ExpyXB(ExpyXB):
    # shape of ExpyXB is (n, )
    return np.sum(np.log(1.+np.reciprocal(ExpyXB)))

def compute_logisticLoss_from_betas_and_yX(betas, yX):
    # shape of betas is (p, )
    # shape of yX is (n, p)
    yXB = yX.dot(betas)
    return compute_logisticLoss_from_yXB(yXB)

def compute_logisticLoss_from_X_y_beta0_betas(X, y, beta0, betas):
    XB = X.dot(betas) + beta0
    yXB = y * XB
    return compute_logisticLoss_from_yXB(yXB)

def convert_y_to_neg_and_pos_1(y):
    y_max, y_min = np.min(y), np.max(y)
    y_transformed = -1 + 2 * (y-y_min)/(y_max-y_min) # convert y to -1 and 1
    return y_transformed

def isEqual_upTo_8decimal(a, b):
    if np.isscalar(a):
        return abs(a - b) < 1e-8
    return np.max(np.abs(a - b)) < 1e-8

def isEqual_upTo_16decimal(a, b):
    if np.isscalar(a):
        return abs(a - b) < 1e-16
    return np.max(np.abs(a - b)) < 1e-16

def insertIntercept_asFirstColOf_X(X):
    n = len(X)
    intercept = np.ones((n, 1))
    X_with_intercept = np.hstack((intercept, X))
    return X_with_intercept

def get_all_product_booleans(sparsity=5):
    # build list of lists:
    all_lists = []
    for i in range(sparsity):
        all_lists.append([0, 1])
    all_products = list(product(*all_lists))
    all_products = [list(elem) for elem in all_products]
    return np.array(all_products)

def download_file_from_google_drive(id, destination):
    # link: https://stackoverflow.com/a/39225272/5040208
    URL = "https://docs.google.com/uc?export=download"

    session = requests.Session()

    response = session.get(URL, params = { 'id' : id , 'confirm': 1 }, stream = True)
    token = get_confirm_token(response)

    if token:
        params = { 'id' : id, 'confirm' : token }
        response = session.get(URL, params = params, stream = True)

    save_response_content(response, destination)

def get_confirm_token(response):
    # link: https://stackoverflow.com/a/39225272/5040208
    for key, value in response.cookies.items():
        if key.startswith('download_warning'):
            return value

    return None

def save_response_content(response, destination):
    # link: https://stackoverflow.com/a/39225272/5040208
    CHUNK_SIZE = 32768

    with open(destination, "wb") as f:
        for chunk in response.iter_content(CHUNK_SIZE):
            if chunk: # filter out keep-alive new chunks
                f.write(chunk)

def check_bounds(bound,  bound_name, num_features):
    if isinstance(bound, (float, int)):
        assert bound >= 0 if bound_name == "ub" else bound <= 0, f"{bound_name} needs to be >= 0" if bound_name == "ub" else f"{bound_name} needs to be <= 0"
    elif isinstance(bound, list):
        bound = np.asarray(bound)
        assert len(bound) == num_features, f"{bound_name}s for the features need to have the same length as the number of features"
        assert np.all(bound >= 0 if bound_name == "ub" else bound <= 0), f"all of {bound_name}s needs to be >= 0" if bound_name == "ub" else f"all of {bound_name}s needs to be <= 0"
    else:
        raise ValueError(f"{bound_name} needs to be a float, int, or list")
    
    return bound


from typing import Type, TypeVar, Optional
import re

T = TypeVar("T", int, float)


def numbers_from_list(
        text: str,
        typ: Type[T],
        minimum: Optional[T] = None,
        maximum: Optional[T] = None,
        enforce_range: Optional[bool] = True) -> tuple[T, ...]:
    text = text.strip()
    if not text:
        return ()

    if re.search(r"(^|[^.])\.\.\.($|[^.])", text):
        return _numbers_from_dots(text, typ, minimum, maximum, enforce_range)
    else:
        if enforce_range:
            return _numbers_from_no_dots(text, typ, minimum, maximum)
        else:
            return _numbers_from_no_dots(text, typ)


def _get_points(numbers: list[str], typ: Type[T]) -> tuple[T, ...]:
    try:
        return tuple(map(typ, numbers))
    except ValueError as exc:
        msg = str(exc)
        raise ValueError(f"invalid value ({msg[msg.rindex(':') + 1:]})") from exc


def _numbers_from_no_dots(
        text: str,
        typ: Type[T],
        minimum: Optional[T] = None,
        maximum: Optional[T] = None) -> tuple[T, ...]:
    points = text.replace("...", " ... ").replace(", ", " ").split()
    steps = tuple(sorted(set(_get_points(points, typ))))
    under = minimum is not None and steps[0] < minimum
    over = maximum is not None and steps[-1] > maximum
    if under and over:
        raise ValueError(f"value must be between {minimum} and {maximum}")
    if under:
        raise ValueError(f"value must be at least {minimum}")
    if over:
        raise ValueError(f"value must be at most {maximum}")
    return steps


def _numbers_from_dots(
        text: str,
        typ: Type[T],
        minimum: Optional[T] = None,
        maximum: Optional[T] = None,
        enforce_range: Optional[bool] = True) -> tuple[T, ...]:
    # many branches are results of many checks and don't degrade readability
    # pylint: disable=too-many-branches
    points = text.replace("...", " ... ").replace(",", " ").split()
    if points.count("...") > 1:
        raise ValueError("multiple '...'.")
    dotind = points.index("...")
    pre = _get_points(points[:dotind], typ)
    post = _get_points(points[dotind + 1:], typ)
    if pre and post and pre[-1] >= post[0]:
        raise ValueError("values before '...' must be smaller than values after.")

    diffs = {y - x for x, y in zip(pre, pre[1:])} \
            | {y - x for x, y in zip(post, post[1:])}
    if not diffs:
        raise ValueError("at least two values are required before or after '...'.")
    diff_of_diffs = max(diffs) - min(diffs)
    if diff_of_diffs > 1e-10:
        raise ValueError("points must be in uniform order.")
    diff = next(iter(diffs))
    if typ is float:
        diff = round(diff, 7)
    if diff <= 0:
        raise ValueError("points must be in increasing order.")

    minpoint = pre[0] if pre else minimum
    maxpoint = post[-1] if post else maximum
    if minpoint is None:
        raise ValueError("minimum value is missing.")
    if maxpoint is None:
        raise ValueError("maximum value is missing.")
    if enforce_range:
        if minimum is not None and minpoint < minimum:
            raise ValueError(f"minimum value is below the minimum {minimum}.")
        if maximum is not None and maxpoint > maximum:
            raise ValueError(f"maximum value is above the maximum {maximum}.")
    steps = (maxpoint - minpoint) // diff
    if (minpoint - maxpoint) % diff > 1e-10:
        if pre and post:
            raise ValueError(
                "the sequence before '...' does not end with the sequence after it.")
        if not pre:
            minpoint = maxpoint - steps * diff
        else:
            maxpoint = minpoint + steps * diff

    if typ is int:
        return tuple(range(minpoint, maxpoint + diff, diff))
    else:
        points = [minpoint + i * diff
                 for i in range(int((maxpoint - minpoint) / diff) + 1)]
        if maxpoint - points[-1] > 1e-10:
            points.append(maxpoint)
        return tuple(points)



"""
This configuration use sphinx-multiproject which builds multiple
Sphinx projects for the Read the Docs. We publish each project at read-the-docs
as orange3 RTD project's subproject. This config file is only required for the
Read the Docs build. Each documentation project can still be built separately
with sphinx-build (make html).

To select a documentation project that the RTD will build, set the PROJECT
environment variable in RTD subprojects to the documentation project name
(e.g. PROJECT=data-mining-library)

To test the documentation build locally run (from doc directory):
```
PROJECT="<project name>" sphinx-build . ./_build
```
More about shpinx-multiproject:
https://sphinx-multiproject.readthedocs.io/en/latest/index.html
"""

# pylint: disable=duplicate-code
extensions = [
    "multiproject",
    "sphinx.ext.autodoc",
    "sphinx.ext.doctest",
    "sphinx.ext.intersphinx",
    "sphinx.ext.todo",
    "sphinx.ext.coverage",
    "sphinx.ext.mathjax",
    "sphinx.ext.ifconfig",
    "sphinx.ext.viewcode",
    "sphinx.ext.autosummary",
    "sphinx.ext.napoleon",
    "recommonmark",
]

# Define the projects that will share this configuration file.
multiproject_projects = {
    "data-mining-library": {
        "path": "data-mining-library/source/"
    },
    "development": {
        "path": "development/source/"
    },
    "visual-programming": {
        "path": "visual-programming/source/"
    },
}
